# Copyright 2020-     Robot Framework Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import inspect
from functools import wraps
from typing import Callable, Tuple

from robot.api import logger  # type: ignore


def attribute_warning(old_args: Tuple[str, str], new_args: Tuple[str, str]):
    def actual_decorator(method):
        @wraps(method)
        def wrapper(*args, **kwargs):
            kw_name = _method_to_keyword(method.__name__)
            for index, old_arg in enumerate(old_args):
                if _is_deprecated_attribute(method, old_arg, args, kwargs):
                    logger.warn(
                        f"Browser library {kw_name} keyword "
                        f"has deprecated {old_arg}, use {new_args[index]}",
                    )
            return method(*args, **kwargs)

        return wrapper

    return actual_decorator


def _method_to_keyword(method: str) -> str:
    keyword = ""
    for word in method.split("_"):
        keyword = f"{keyword} {word.capitalize()}"
    return keyword.strip()


def _is_deprecated_attribute(method: Callable, deprecated_arg, args, kwargs):
    if not deprecated_arg:
        return False
    args = list(args)
    args.pop(0)
    argspec = inspect.getfullargspec(method)
    argspec_args = argspec.args
    argspec_args.pop(0)
    deprecated = False
    if not args and not kwargs:
        deprecated = False
    if deprecated_arg in kwargs:
        deprecated = True
    for index, argspec_arg in enumerate(argspec_args):
        if argspec_arg in deprecated_arg:
            if len(args) == index + 1:
                deprecated = True
    return deprecated



# Copyright 2020-     Robot Framework Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import argparse
import os
import platform
import shutil
import subprocess
import sys
import traceback
from pathlib import Path
from subprocess import DEVNULL, PIPE, STDOUT, CalledProcessError, Popen
from typing import TextIO

INSTALLATION_DIR = Path(__file__).parent / "wrapper"
NODE_MODULES = INSTALLATION_DIR / "node_modules"
# This is required because weirdly windows doesn't have `npm` in PATH without shell=True.
# But shell=True breaks our linux CI
SHELL = True if platform.platform().startswith("Windows") else False


def _write_marker(file, first=False):
    if not first:
        file.write("\n")
    file.write("=" * 110)
    file.write("\n")


def rfbrowser_init(skip_browser_install: bool):
    current_folder = Path(__file__).resolve().parent
    install_log = current_folder / "rfborwser.log"
    with open(install_log, "w") as install_file:
        _write_marker(install_file, True)
        try:
            _rfbrowser_init(skip_browser_install, install_file)
            _write_marker(install_file)
        except Exception as error:
            _write_marker(install_file)
            install_file.write(traceback.format_exc())
            _python_info(install_file)
            _node_info(install_file)
            _log_install_dir(install_file)
            raise error


def _log_install_dir(install_file):
    install_file.write(
        f"Installation directory `{INSTALLATION_DIR}` does not contain the required files"
        "\nPrinting contents:\n"
    )
    for line in _walk_install_dir():
        install_file.write(line)
    _write_marker(install_file)


def _node_info(install_file):
    process = subprocess.run(
        ["npm", "-v"], stdout=subprocess.PIPE, stderr=subprocess.STDOUT, shell=SHELL
    )
    install_file.write("npm version is:\n")
    install_file.write(process.stdout.decode("UTF-8"))
    _write_marker(install_file)


def _python_info(install_file):
    _write_marker(install_file)
    python_version = (
        f"{sys.version_info.major}.{sys.version_info.minor}.{sys.version_info.micro}"
    )
    install_file.write(f"Used Python is: {sys.executable}\nVersion: {python_version}")
    _write_marker(install_file)
    install_file.write("pip freeze output:\n\n")
    process = subprocess.run(
        [sys.executable, "-m", "pip", "freeze"],
        stderr=subprocess.STDOUT,
        stdout=subprocess.PIPE,
        timeout=60,
    )
    install_file.write(process.stdout.decode("UTF-8"))
    _write_marker(install_file)


def _walk_install_dir():
    lines = []
    for root, _dirs, files in os.walk(INSTALLATION_DIR):
        level = root.replace(INSTALLATION_DIR.__str__(), "").count(os.sep)
        indent = " " * 4 * (level)
        lines.append("{}{}/\n".format(indent, os.path.basename(root)))
        subindent = " " * 4 * (level + 1)
        for file in files:
            lines.append("{}{}\n".format(subindent, file))
    return lines


def _rfbrowser_init(skip_browser_install: bool, log_file: TextIO):
    print("Installing node dependencies...")
    if not (INSTALLATION_DIR / "package.json").is_file():
        print(
            f"Installation directory `{INSTALLATION_DIR}` does not contain the required package.json "
            + "\nPrinting contents:\n"
        )
        for line in _walk_install_dir():
            print(line)
        raise RuntimeError("Could not find robotframework-browser's package.json")
    if not os.access(INSTALLATION_DIR, os.W_OK):
        sys.tracebacklimit = 0
        raise RuntimeError(
            f"`rfbrowser init` needs write permissions to {INSTALLATION_DIR}"
        )

    print(f"Installing rfbrowser node dependencies at {INSTALLATION_DIR}")

    try:
        subprocess.run(["npm", "-v"], stdout=DEVNULL, check=True, shell=SHELL)
    except (CalledProcessError, FileNotFoundError, PermissionError) as exception:
        print(
            "Couldn't execute npm. Please ensure you have node.js and npm installed and in PATH."
            "See https://nodejs.org/ for documentation"
        )
        raise exception

    if skip_browser_install:
        os.environ["PLAYWRIGHT_SKIP_BROWSER_DOWNLOAD"] = "1"
    else:
        if not os.environ.get("PLAYWRIGHT_BROWSERS_PATH"):
            os.environ["PLAYWRIGHT_BROWSERS_PATH"] = "0"

    process = Popen(
        "npm ci --production",
        shell=True,
        cwd=INSTALLATION_DIR,
        stdout=PIPE,
        stderr=STDOUT,
    )

    while process.poll() is None:
        if process.stdout:
            output = process.stdout.readline().decode("UTF-8")
            log_file.write(output)
            print(output)

    if process.returncode != 0:
        raise RuntimeError(
            "Problem installing node dependencies."
            + f"Node process returned with exit status {process.returncode}"
        )

    print("rfbrowser init completed")


def rfbrowser_clean_node():
    if not NODE_MODULES.is_dir():
        print(f"Could not find {NODE_MODULES}, nothing to delete.")
        return
    print("Delete library node dependencies...")
    shutil.rmtree(NODE_MODULES)


def show_trace(file: str):
    print(f"Opening file: {file}")
    playwright = NODE_MODULES / "playwright"
    local_browsers = playwright / ".local-browsers"
    env = os.environ.copy()
    env["PLAYWRIGHT_BROWSERS_PATH"] = str(local_browsers)
    trace_arguments = [
        "npx",
        "playwright",
        "show-trace",
        file,
    ]
    subprocess.run(trace_arguments, env=env, shell=SHELL)


# Based on: https://stackoverflow.com/questions/3853722/how-to-insert-newlines-on-argparse-help-text
class SmartFormatter(argparse.HelpFormatter):
    def _split_lines(self, text, width):
        if text.startswith("Possible commands are:"):
            parts = []
            for part in text.splitlines():
                part = argparse.HelpFormatter._split_lines(self, part, width)
                parts.extend(part if part else "\n")
            return parts
        return argparse.HelpFormatter._split_lines(self, text, width)


def run():
    parser = argparse.ArgumentParser(
        description="Robot Framework Browser library command line tool. If there is error during command, debug "
        "information is saved to <install dir>/Browser/rfborwser.log file.",
        formatter_class=SmartFormatter,
    )
    parser.add_argument(
        "command",
        help=(
            "Possible commands are:\ninit\nclean-node\nshow-trace\n\ninit command will install the required node "
            "dependencies. init command is needed when library is installed or updated.\n\nclean-node is used to delete"
            "node side dependencies and installed browser binaries from the library default installation location. "
            "When upgrading browser library, it is recommended to clean old node side binaries after upgrading the "
            "Python side. Example:\n1) pip install -U robotframework-browser\n2) rfbrowser clean-node\n3)rfbrowser "
            "init.\nRun rfbrowser clean-node command also before uninstalling the library with pip. This makes sure "
            "that playwright browser binaries are not left in the disk after the pip uninstall command."
            "\n\nshow-trace command will start the Playwright trace viewer tool.\n\nSee the each command argument "
            "group for more details what (optional) arguments that command supports."
        ),
        type=str,
    )
    install = parser.add_argument_group("init options")
    install.add_argument(
        "--skip-browsers",
        help="If defined skips the Playwright browser installation. Argument is optional",
        default=False,
        action="store_true",
    )
    trace = parser.add_argument_group("show-trace options")
    trace.add_argument(
        "--file",
        "-F",
        help=(
            "Full path to trace zip file. See New Context keyword for more details how to "
            "create trace file. Argument is mandatory."
        ),
        default=False,
    )
    args = parser.parse_args()
    command = args.command.lower()
    if command == "init":
        rfbrowser_init(args.skip_browsers)
    elif command == "clean-node":
        rfbrowser_clean_node()
    elif command == "show-trace":
        if not args.file:
            raise Exception("show-trace needs also --file argument")
        show_trace(args.file)
    else:
        raise Exception(
            f"Command should be init, clean-node or show-trace, but it was {command}"
        )


if __name__ == "__main__":
    run()



# Copyright 2020-     Robot Framework Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import contextlib
import inspect
import os
import socket
import subprocess
from os import PathLike
from typing import Any, Tuple

from robot.libraries.BuiltIn import BuiltIn  # type: ignore

get_variable_value = BuiltIn().get_variable_value


def find_free_port() -> int:
    with contextlib.closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
        s.bind(("", 0))
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        return s.getsockname()[1]


def spawn_node_process(output_dir: PathLike) -> Tuple[subprocess.Popen, str]:
    """
    Spawn an rfbrowser node process, that can be shared between library instances.

    Usage example:

    rc = 1
    background_process, port = spawn_node_process(ATEST_OUTPUT / "playwright-log.txt")
    try:
        os.environ["ROBOT_FRAMEWORK_BROWSER_NODE_PORT"] = port
        rc = _run_pabot(args)
    finally:
        background_process.kill()


    """
    logfile = open(output_dir, "w")
    os.environ["DEBUG"] = "pw:api"
    os.environ["PLAYWRIGHT_BROWSERS_PATH"] = "0"
    port = str(find_free_port())
    process = subprocess.Popen(
        [
            "node",
            "Browser/wrapper/index.js",
            port,
        ],
        stdout=logfile,
        stderr=subprocess.STDOUT,
    )
    return process, port


def is_same_keyword(first: str, second: str) -> bool:
    if isinstance(first, str) and isinstance(second, str):
        return get_normalized_keyword(first) == get_normalized_keyword(second)
    else:
        return False


def get_normalized_keyword(keyword: str) -> str:
    return keyword.lower().replace(" ", "_")


def keyword(name: Any = None, tags: Tuple = (), types: Tuple = ()):
    if inspect.isroutine(name):
        return keyword()(name)

    def decorator(func):
        func.robot_name = name
        func.robot_tags = tags
        func.robot_types = types
        return func

    return decorator


def type_converter(argument: Any) -> str:
    return type(argument).__name__.lower()



# Copyright 2020-     Robot Framework Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from ..base import LibraryComponent
from ..utils import keyword, logger


class StrictMode(LibraryComponent):
    @keyword(tags=("Setter", "BrowserControl"))
    def set_strict_mode(self, mode: bool):
        """Controls library strict mode.

        When set to ``True``, keywords that are searching elements will use Playwright
        [https://playwright.dev/docs/api/class-page#page-query-selector|strict mode].
        Keyword changes library strict mode value and keyword also return the previous
        strict mode value.

        Example:
        | ${old_mode} =      Set Strict Mode    False
        | Get Text           //input            # Does not fail if selector points to one or more elements
        | Set Strict Mode    ${old_mode}
        """
        old_mode = self.strict_mode
        self.strict_mode = mode
        logger.debug(f"Old mode was {old_mode}")
        return old_mode



# Copyright 2020-     Robot Framework Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
import os
import re
import shutil
import string
import sys
import time
import types
from concurrent.futures._base import Future
from datetime import timedelta
from pathlib import Path
from typing import Dict, List, Optional, Set, Union

from assertionengine import AssertionOperator, Formatter
from overrides import overrides
from robot.errors import DataError  # type: ignore
from robot.libraries.BuiltIn import EXECUTION_CONTEXTS, BuiltIn  # type: ignore
from robot.result.model import TestCase as TestCaseResult  # type: ignore
from robot.running.arguments import PythonArgumentParser  # type: ignore
from robot.running.model import TestCase as TestCaseRunning  # type: ignore
from robot.utils import secs_to_timestr, timestr_to_secs  # type: ignore
from robotlibcore import DynamicCore  # type: ignore

from .base import ContextCache, LibraryComponent
from .generated.playwright_pb2 import Request
from .keywords import (
    Control,
    Cookie,
    Devices,
    Evaluation,
    Getters,
    Interaction,
    Network,
    PlaywrightState,
    Promises,
    RunOnFailureKeywords,
    StrictMode,
    Waiter,
    WebAppState,
)
from .keywords.crawling import Crawling
from .playwright import Playwright
from .utils import AutoClosingLevel, get_normalized_keyword, is_falsy, keyword, logger

# Importing this directly from .utils break the stub type checks
from .utils.data_types import DelayedKeyword, HighLightElement, SupportedBrowsers
from .version import __version__ as VERSION


class Browser(DynamicCore):
    """Browser library is a browser automation library for Robot Framework.

    This is the keyword documentation for Browser library. For information
    about installation, support, and more please visit the
    [https://github.com/MarketSquare/robotframework-playwright|project pages].
    For more information about Robot Framework itself, see [https://robotframework.org|robotframework.org].

    Browser library uses
    [https://github.com/microsoft/playwright|Playwright Node module]
    to automate [https://www.chromium.org/Home|Chromium],
    [https://www.mozilla.org/en-US/firefox/new/|Firefox]
    and [https://webkit.org/|WebKit] with a single library.


    == Table of contents ==

    %TOC%

    = Browser, Context and Page =

    Browser library works with three different layers that build on each other:
    *Browser*, *Context* and *Page*.


    == Browsers ==

    A *browser* can be started with one of the three
    different engines Chromium, Firefox or Webkit.

    === Supported Browsers ===

    |   Browser     | Browser with this engine                          |
    | ``chromium``  | Google Chrome, Microsoft Edge (since 2020), Opera |
    | ``firefox``   | Mozilla Firefox                                   |
    | ``webkit``    | Apple Safari, Mail, AppStore on MacOS and iOS     |

    Since [https://github.com/microsoft/playwright|Playwright] comes with a pack of builtin
    binaries for all browsers, no additional drivers e.g. geckodriver are needed.

    All these browsers that cover more than 85% of the world wide used browsers,
    can be tested on Windows, Linux and MacOS.
    There is no need for dedicated machines anymore.

    A browser process is started ``headless`` (without a GUI) by default.
    Run `New Browser` with specified arguments if a browser with a GUI is requested
    or if a proxy has to be configured.
    A browser process can contain several contexts.


    == Contexts ==

    A *context* corresponds to set of independent incognito pages in a browser
    that share cookies, sessions or profile settings. Pages in two separate
    contexts do not share cookies, sessions or profile settings.
    Compared to Selenium, these do *not* require their own browser process.
    To get a clean environment a test can just open a new context.
    Due to this new independent browser sessions can be opened with
    Robot Framework Browser about 10 times faster than with Selenium by
    just opening a `New Context` within the opened browser.

    To make pages in the same suite share state, use the same context by opening the
    context with `New Context` on suite setup.

    The context layer is useful e.g. for testing different users sessions on the
    same webpage without opening a whole new browser context.
    Contexts can also have detailed configurations, such as geo-location, language settings,
    the viewport size or color scheme.
    Contexts do also support http credentials to be set, so that basic authentication
    can also be tested. To be able to download files within the test,
    the ``acceptDownloads`` argument must be set to ``True`` in `New Context` keyword.
    A context can contain different pages.


    == Pages ==

    A *page* does contain the content of the loaded web site and has a browsing history.
    Pages and browser tabs are the same.

    Typical usage could be:
    | *** Test Cases ***
    | Starting a browser with a page
    |     New Browser    chromium    headless=false
    |     New Context    viewport={'width': 1920, 'height': 1080}
    |     New Page       https://marketsquare.github.io/robotframework-browser/Browser.html
    |     Get Title      ==    Browser

    The `Open Browser` keyword opens a new browser, a new context and a new page.
    This keyword is useful for quick experiments or debugging sessions.

    When a `New Page` is called without an open browser, `New Browser`
    and `New Context` are executed with default values first.

    Each Browser, Context and Page has a unique ID with which they can be addressed.
    A full catalog of what is open can be received by `Get Browser Catalog` as dictionary.

    = Automatic page and context closing =

    %AUTO_CLOSING_LEVEL%

    = Finding elements =

    All keywords in the library that need to interact with an element
    on a web page take an argument typically named ``selector`` that specifies
    how to find the element. Keywords can find elements with strict mode. If
    strict mode is true and locator finds multiple elements from the page, keyword
    will fail. If keyword finds one element, keyword does not fail because of
    strict mode. If strict mode is false, keyword does not fail if selector points
    many elements. Strict mode is enabled by default, but can be changed in library
    `importing` or `Set Strict Mode` keyword. Keyword documentation states if keyword
    uses strict mode. If keyword does not state that is used strict mode, then strict
    mode is not applied for the keyword. For more details, see Playwright
    [https://playwright.dev/docs/api/class-page#page-query-selector|strict documentation].

    Selector strategies that are supported by default are listed in the table
    below.

    | = Strategy = |     = Match based on =     |         = Example =                |
    | ``css``      | CSS selector.              | ``css=.class > \\#login_btn``      |
    | ``xpath``    | XPath expression.          | ``xpath=//input[@id="login_btn"]`` |
    | ``text``     | Browser text engine.       | ``text=Login``                     |
    | ``id``       | Element ID Attribute.      | ``id=login_btn``                   |

    CSS Selectors can also be recorded with `Record selector` keyword.

    == Explicit Selector Strategy ==

    The explicit selector strategy is specified with a prefix using syntax
    ``strategy=value``. Spaces around the separator are ignored, so
    ``css=foo``, ``css= foo`` and ``css = foo`` are all equivalent.


    == Implicit Selector Strategy ==

    *The default selector strategy is `css`.*

    If selector does not contain one of the know explicit selector strategies, it is
    assumed to contain css selector.

    Selectors that are starting with ``//`` or ``..`` are considered as xpath selectors.

    Selectors that are in quotes are considered as text selectors.

    Examples:

    | # CSS selectors are default.
    | `Click`  span > button.some_class         # This is equivalent
    | `Click`  css=span > button.some_class     # to this.
    |
    | # // or .. leads to xpath selector strategy
    | `Click`  //span/button[@class="some_class"]
    | `Click`  xpath=//span/button[@class="some_class"]
    |
    | # "text" in quotes leads to exact text selector strategy
    | `Click`  "Login"
    | `Click`  text="Login"


    == CSS ==

    As written before, the default selector strategy is `css`. See
    [https://developer.mozilla.org/en-US/docs/Web/CSS/CSS_Selectors | css selector]
    for more information.

    Any malformed selector not starting with ``//`` or ``..`` nor starting and ending
    with a quote is assumed to be a css selector.

    Note that ``#`` is a comment character in [https://robotframework.org/robotframework/latest/RobotFrameworkUserGuide.html#ignored-data | Robot Framework syntax] and needs to be
    escaped like ``\\#`` to work as a [https://developer.mozilla.org/en-US/docs/Web/CSS/ID_selectors | css ID selector].

    Examples:
    | `Click`  span > button.some_class
    | `Get Text`  \\#username_field  ==  George


    == XPath ==

    XPath engine is equivalent to [https://developer.mozilla.org/en/docs/Web/API/Document/evaluate|Document.evaluate].
    Example: ``xpath=//html/body//span[text()="Hello World"]``.

    Malformed selector starting with ``//`` or ``..`` is assumed to be an xpath selector.
    For example, ``//html/body`` is converted to ``xpath=//html/body``. More
    examples are displayed in `Examples`.

    Note that xpath does not pierce [https://developer.mozilla.org/en-US/docs/Web/Web_Components/Using_shadow_DOM|shadow_roots].


    == Text ==

    Text engine finds an element that contains a text node with the passed text.
    For example, ``Click    text=Login`` clicks on a login button, and
    ``Wait For Elements State   text="lazy loaded text"`` waits for the "lazy loaded text"
    to appear in the page.

    Text engine finds fields based on their labels in text inserting keywords.

    Malformed selector starting and ending with a quote (either ``"`` or ``'``) is assumed
    to be a text selector. For example, ``Click    "Login"`` is converted to ``Click    text="Login"``.
    Be aware that these leads to exact matches only!
    More examples are displayed in `Examples`.


    === Insensitive match ===

    By default, the match is case-insensitive, ignores leading/trailing whitespace and
    searches for a substring. This means ``text= Login`` matches
    ``<button>Button loGIN (click me)</button>``.

    === Exact match ===

    Text body can be escaped with single or double quotes for precise matching,
    insisting on exact match, including specified whitespace and case.
    This means ``text="Login "`` will only match ``<button>Login </button>`` with exactly
    one space after "Login". Quoted text follows the usual escaping rules, e.g.
    use ``\\"`` to escape double quote in a double-quoted string: ``text="foo\\"bar"``.

    === RegEx ===

    Text body can also be a JavaScript-like regex wrapped in / symbols.
    This means ``text=/^hello .*!$/i`` or ``text=/^Hello .*!$/`` will match ``<span>Hello Peter Parker!</span>``
    with any name after ``Hello``, ending with ``!``.
    The first one flagged with ``i`` for case-insensitive.
    See [https://regex101.com/|https://regex101.com] for more information about RegEx.

    === Button and Submit Values ===

    Input elements of the type button and submit are rendered with their value as text,
    and text engine finds them. For example, ``text=Login`` matches
    ``<input type=button value="Login">``.

    == Cascaded selector syntax ==

    Browser library supports the same selector strategies as the underlying
    Playwright node module: xpath, css, id and text. The strategy can either
    be explicitly specified with a prefix or the strategy can be implicit.

    A major advantage of Browser is, that multiple selector engines can be used
    within one selector. It is possible to mix XPath, CSS and Text selectors while
    selecting a single element.

    Selectors are strings that consists of one or more clauses separated by
    ``>>`` token, e.g. ``clause1 >> clause2 >> clause3``. When multiple clauses
    are present, next one is queried relative to the previous one's result.
    Browser library supports concatenation of different selectors separated by ``>>``.

    For example:
    | `Highlight Elements`    "Hello" >> ../.. >> .select_button
    | `Highlight Elements`    text=Hello >> xpath=../.. >> css=.select_button

    Each clause contains a selector engine name and selector body, e.g.
    ``engine=body``. Here ``engine`` is one of the supported engines (e.g. css or
    a custom one). Selector ``body`` follows the format of the particular engine,
    e.g. for css engine it should be a [https://developer.mozilla.org/en-US/docs/Web/CSS/CSS_Selectors | css selector].
    Body format is assumed to ignore leading and trailing white spaces,
    so that extra whitespace can be added for readability. If selector
    engine needs to include ``>>`` in the body, it should be escaped
    inside a string to not be confused with clause separator,
    e.g. ``text="some >> text"``.

    Selector engine name can be prefixed with ``*`` to capture element that
    matches the particular clause instead of the last one. For example,
    ``css=article >> text=Hello`` captures the element with the text ``Hello``,
    and ``*css=article >> text=Hello`` (note the *) captures the article element
    that contains some element with the text Hello.

    For convenience, selectors in the wrong format are heuristically converted
    to the right format. See `Implicit Selector Strategy`

    == Examples ==
    | # queries 'div' css selector
    | Get Element    css=div
    |
    | # queries '//html/body/div' xpath selector
    | Get Element    //html/body/div
    |
    | # queries '"foo"' text selector
    | Get Element    text=foo
    |
    | # queries 'span' css selector inside the result of '//html/body/div' xpath selector
    | Get Element    xpath=//html/body/div >> css=span
    |
    | # converted to 'css=div'
    | Get Element    div
    |
    | # converted to 'xpath=//html/body/div'
    | Get Element    //html/body/div
    |
    | # converted to 'text="foo"'
    | Get Element    "foo"
    |
    | # queries the div element of every 2nd span element inside an element with the id foo
    | Get Element    \\#foo >> css=span:nth-child(2n+1) >> div
    | Get Element    id=foo >> css=span:nth-child(2n+1) >> div

    Be aware that using ``#`` as a starting character in Robot Framework would be interpreted as comment.
    Due to that fact a ``#id`` must be escaped as ``\\#id``.

    == Frames ==

    By default, selector chains do not cross frame boundaries. It means that a
    simple CSS selector is not able to select and element located inside an iframe
    or a frameset. For this use case, there is a special selector ``>>>`` which can
    be used to combine a selector for the frame and a selector for an element
    inside a frame.

    Given this simple pseudo html snippet:
    | <iframe id="iframe" src="src.html">
    |   #document
    |     <!DOCTYPE html>
    |     <html>
    |       <head></head>
    |       <body>
    |         <button id="btn">Click Me</button>
    |       </body>
    |     </html>
    | </iframe>

    Here's a keyword call that clicks the button inside the frame.

    | Click   id=iframe >>> id=btn

    The selectors on the left and right side of ``>>>`` can be any valid selectors.
    The selector clause directly before the frame opener ``>>>`` must select the frame element.
    Frame selection is the only place where Library modifies the selector, as explained in above.
    In all cases, library does not alter selector in any way, instead it is passed as is to
    Playwright side.

    == WebComponents and Shadow DOM ==

    Playwright and so also Browser are able to do automatic piercing of Shadow DOMs
    and therefore are the best automation technology when working with WebComponents.

    Also other technologies claim that they can handle
    [https://developer.mozilla.org/en-US/docs/Web/Web_Components/Using_shadow_DOM|Shadow DOM and Web Components].
    However, non of them do pierce shadow roots automatically,
    which may be inconvenient when working with Shadow DOM and Web Components.

    For that reason, css engine pierces shadow roots. More specifically, every
    [https://developer.mozilla.org/en-US/docs/Web/CSS/Descendant_combinator|Descendant combinator]
    pierces an arbitrary number of open shadow roots, including the implicit descendant combinator
    at the start of the selector.

    That means, it is not necessary to select each shadow host, open its shadow root and
    select the next shadow host until you reach the element that should be controlled.

    === CSS:light ===

    ``css:light`` engine is equivalent to [https://developer.mozilla.org/en/docs/Web/API/Document/querySelector | Document.querySelector]
    and behaves according to the CSS spec.
    However, it does not pierce shadow roots.

    ``css`` engine first searches for elements in the light dom in the iteration order,
    and then recursively inside open shadow roots in the iteration order. It does not
    search inside closed shadow roots or iframes.

    Examples:

    | <article>
    |   <div>In the light dom</div>
    |   <div slot='myslot'>In the light dom, but goes into the shadow slot</div>
    |   <open mode shadow root>
    |       <div class='in-the-shadow'>
    |           <span class='content'>
    |               In the shadow dom
    |               <open mode shadow root>
    |                   <li id='target'>Deep in the shadow</li>
    |               </open mode shadow root>
    |           </span>
    |       </div>
    |       <slot name='myslot'></slot>
    |   </open mode shadow root>
    | </article>

    Note that ``<open mode shadow root>`` is not an html element, but rather a shadow root
    created with ``element.attachShadow({mode: 'open'})``.

    - Both ``"css=article div"`` and ``"css:light=article div"`` match the first ``<div>In the light dom</div>``.
    - Both ``"css=article > div"`` and ``"css:light=article > div"`` match two ``div`` elements that are direct children of the ``article``.
    - ``"css=article .in-the-shadow"`` matches the ``<div class='in-the-shadow'>``, piercing the shadow root, while ``"css:light=article .in-the-shadow"`` does not match anything.
    - ``"css:light=article div > span"`` does not match anything, because both light-dom ``div`` elements do not contain a ``span``.
    - ``"css=article div > span"`` matches the ``<span class='content'>``, piercing the shadow root.
    - ``"css=article > .in-the-shadow"`` does not match anything, because ``<div class='in-the-shadow'>`` is not a direct child of ``article``
    - ``"css:light=article > .in-the-shadow"`` does not match anything.
    - ``"css=article li#target"`` matches the ``<li id='target'>Deep in the shadow</li>``, piercing two shadow roots.

    === text:light ===

    ``text`` engine open pierces shadow roots similarly to ``css``, while ``text:light`` does not.
    Text engine first searches for elements in the light dom in the iteration order, and then
    recursively inside open shadow roots in the iteration order. It does not search inside
    closed shadow roots or iframes.

    === id, data-testid, data-test-id, data-test and their :light counterparts ===

    Attribute engines are selecting based on the corresponding attribute value.
    For example: ``data-test-id=foo`` is equivalent to ``css=[data-test-id="foo"]``,
    and ``id:light=foo`` is equivalent to ``css:light=[id="foo"]``.

    == Element reference syntax ==

    It is possible to get a reference to a Locator by using `Get Element` and `Get Elements` keywords.
    Keywords do not save reference to an element in the HTML document, instead it saves reference to a Playwright
    [https://playwright.dev/docs/api/class-locator|Locator]. In nutshell Locator captures the logic of how to
    retrieve that element from the page. Each time action is performed, locator re-searches the elements
    in the page. This reference can be used as a *first* part of a selector by using a special selector
    syntax `element=`. like this:

    | ${ref}=    Get Element    .some_class
    |            Click          ${ref} >> .some_child     # Locator searches an element from the page.
    |            Click          ${ref} >> .other_child    # Locator searches again an element from the page.

    The `.some_child` and `.other_child` selectors in the example are relative to the element referenced
    by ${ref}. Please note that frame piercing is not possible with element reference.

    = Assertions =

    Keywords that accept arguments ``assertion_operator`` <`AssertionOperator`> and ``assertion_expected``
    can optionally assert that a specified condition holds. Keywords will return the value even when the
    assertion is performed by the keyword.

    Assert will retry and fail only after a specified timeout.
    See `Importing` and ``retry_assertions_for`` (default is 1 second) for configuring this timeout.


    %ASSERTION_TABLE%

    By default, keywords will provide an error message if an assertion fails.
    Default error message can be overwritten with a ``message`` argument.
    The ``message`` argument accepts `{value}`, `{value_type}`, `{expected}` and
    `{expected_type}` [https://docs.python.org/3/library/stdtypes.html#str.format|format]
    options.
    The `{value}` is value returned by the keyword and the `{expected}`
    is expected value defined by the user, usually value in the
    ``assertion_expected`` argument. The `{value_type}` and
    `{expected_type}` are the type definitions from `{value}` and `{expected}`
    arguments. In similar fashion as Python
    [https://docs.python.org/3/library/functions.html#type|type] returns type definition.
    Assertions will retry until ``timeout`` has expired if they do not pass.

    The assertion ``assertion_expected`` value is not converted by the library and
    is used as is. Therefore when assertion is made, the ``assertion_expected``
    argument value and value returned the keyword must have same type. If types
    are not same, assertion will fail. Example `Get Text` always returns a string
    and has to be compared with a string, even the returned value might look like
    a number.

    Other Keywords have other specific types they return.
    `Get Element Count` always returns an integer.
    `Get Bounding Box` and `Get Viewport Size` can be filtered.
    They return a dictionary without filter and a number when filtered.
    These Keywords do automatic conversion for the expected value if a number is returned.

    * < less or greater > With Strings*
    Comparisons of strings with ``greater than`` or ``less than`` compares each character,
    starting from 0 regarding where it stands in the code page.
    Example: ``A < Z``, ``Z < a``, ``ac < dc`
    It does never compare the length of elements. Neither lists nor strings.
    The comparison stops at the first character that is different.
    Examples: ``'abcde' < 'abd'``, ``'100.000' < '2'``
    In Python 3 and therefore also in Browser it is not possible to compare numbers
    with strings with a greater or less operator.
    On keywords that return numbers, the given expected value is automatically
    converted to a number before comparison.


    The getters `Get Page State` and `Get Browser Catalog` return a dictionary. Values of the dictionary can directly asserted.
    Pay attention of possible types because they are evaluated in Python. For example:

    | Get Page State    validate    2020 >= value['year']                     # Comparison of numbers
    | Get Page State    validate    "IMPORTANT MESSAGE!" == value['message']  # Comparison of strings

    == The 'then' or 'evaluate' closure ==

    Keywords that accept arguments ``assertion_operator`` and ``assertion_expected``
    can optionally also use ``then`` or ``evaluate`` closure to modify the returned value with
    BuiltIn Evaluate. Actual value can be accessed with ``value``.

    For example ``Get Title  then  'TITLE: '+value``.
    See
    [https://robotframework.org/robotframework/latest/libraries/BuiltIn.html#Evaluating%20expressions|
    Builtin Evaluating expressions]
    for more info on the syntax.

    == Examples ==

    | # *Keyword*    *Selector*                    *Key*        *Assertion Operator*    *Assertion Expected*
    | Get Title                                           equal                 Page Title
    | Get Title                                           ^=                    Page
    | Get Style    //*[@id="div-element"]      width      >                     100
    | Get Title                                           matches               \\\\w+\\\\s\\\\w+
    | Get Title                                           validate              value == "Login Page"
    | Get Title                                           evaluate              value if value == "some value" else "something else"


    = Implicit waiting =

    Browser library and Playwright have many mechanisms to help in waiting for elements.
    Playwright will auto-wait before performing actions on elements.
    Please see [https://playwright.dev/docs/actionability/ | Auto-waiting on Playwright documentation]
    for more information.

    On top of Playwright auto-waiting Browser assertions will wait and retry
    for specified time before failing any `Assertions`.
    Time is specified in Browser library initialization with ``retry_assertions_for``.

    Browser library also includes explicit waiting keywords such as `Wait for Elements State`
    if more control for waiting is needed.

    = Experimental: Re-using same node process =

    Browser library integrated nodejs and python. NodeJS side can be also executed as a standalone process.
    Browser libraries running on the same machine can talk to that instead of starting new node processes.
    This can speed execution when running tests parallel.
    To start node side run on the directory when Browser package is
    ``PLAYWRIGHT_BROWSERS_PATH=0 node Browser/wrapper/index.js PORT``.
    ``PORT`` is port you want to use for the node process.
    To execute tests then with pabot for example do ``ROBOT_FRAMEWORK_BROWSER_NODE_PORT=PORT pabot ..``.

    = Extending Browser library with a JavaScript module =

    Browser library can be extended with JavaScript. Module must be in CommonJS format that Node.js uses.
    You can translate your ES6 module to Node.js CommonJS style with Babel. Many other languages
    can be also translated to modules that can be used from Node.js. For example TypeScript, PureScript and
    ClojureScript just to mention few.

    | async function myGoToKeyword(url, args, page, logger, playwright) {
    |   logger(args.toString())
    |   playwright.coolNewFeature()
    |   return await page.goto(url);
    | }

    Functions can contain any number of arguments and arguments may have default values.

    There are some reserved arguments that are not accessible from Robot Framework side.
    They are injected to the function if they are in the arguments:

    ``page``: [https://playwright.dev/docs/api/class-page|the playwright Page object].

    ``args``: the rest of values from Robot Framework keyword call ``*args``.

    ``logger``: callback function that takes strings as arguments and writes them to robot log. Can be called multiple times.

    ``playwright``: playwright module (* from 'playwright'). Useful for integrating with Playwright features that Browser library doesn't support with it's own keywords. [https://playwright.dev/docs/api/class-playwright| API docs]

    also argument name ``self`` can not be used.

    == Example module.js ==

    | async function myGoToKeyword(pageUrl, page) {
    |   await page.goto(pageUrl);
    |   return await page.title();
    | }
    | exports.__esModule = true;
    | exports.myGoToKeyword = myGoToKeyword;

    == Example Robot Framework side ==

    | *** Settings ***
    | Library   Browser  jsextension=${CURDIR}/module.js
    |
    | *** Test Cases ***
    | Hello
    |   New Page
    |   ${title}=  myGoToKeyword  https://playwright.dev
    |   Should be equal  ${title}  Playwright

    Also selector syntax can be extended with a custom selector using a js module

    == Example module keyword for custom selector registering ==

    | async function registerMySelector(playwright) {
    | playwright.selectors.register("myselector", () => ({
    |    // Returns the first element matching given selector in the root's subtree.
    |    query(root, selector) {
    |       return root.querySelector(`a[data-title="${selector}"]`);
    |     },
    |
    |     // Returns all elements matching given selector in the root's subtree.
    |     queryAll(root, selector) {
    |       return Array.from(root.querySelectorAll(`a[data-title="${selector}"]`));
    |     }
    | }));
    | return 1;
    | }
    | exports.__esModule = true;
    | exports.registerMySelector = registerMySelector;
    """

    ROBOT_LIBRARY_VERSION = VERSION
    ROBOT_LISTENER_API_VERSION = 3
    ROBOT_LIBRARY_LISTENER: "Browser"
    ROBOT_LIBRARY_SCOPE = "GLOBAL"
    _context_cache = ContextCache()
    _suite_cleanup_done = False

    def __init__(
        self,
        timeout: timedelta = timedelta(seconds=10),
        enable_playwright_debug: bool = False,
        auto_closing_level: AutoClosingLevel = AutoClosingLevel.TEST,
        retry_assertions_for: timedelta = timedelta(seconds=1),
        run_on_failure: str = "Take Screenshot  fail-screenshot-{index}",
        external_browser_executable: Optional[Dict[SupportedBrowsers, str]] = None,
        jsextension: Optional[str] = None,
        enable_presenter_mode: Union[HighLightElement, bool] = False,
        playwright_process_port: Optional[int] = None,
        strict: bool = True,
    ):
        """Browser library can be taken into use with optional arguments:

        - ``timeout`` <str>
          Timeout for keywords that operate on elements. The keywords will wait
          for this time for the element to appear into the page. Defaults to "10s" => 10 seconds.
        - ``enable_playwright_debug`` <bool>
          Enable low level debug information from the playwright tool. Mainly
          Useful for the library developers and for debugging purposes.
        - ``auto_closing_level`` < ``TEST`` | ``SUITE`` | ``MANUAL`` >
          Configure context and page automatic closing. Default is ``TEST``,
          for more details, see `AutoClosingLevel`
        - ``retry_assertions_for`` <str>
          Timeout for retrying assertions on keywords before failing the keywords.
          This timeout starts counting from the first failure.
          Global ``timeout`` will still be in effect.
          This allows stopping execution faster to assertion failure when element is found fast.
        - ``run_on_failure`` <str>
          Sets the keyword to execute in case of a failing Browser keyword.
          It can be the name of any keyword. If the keyword has arguments those must be separated with
          two spaces for example ``My keyword \\ arg1 \\ arg2``.
          If no extra action should be done after a failure, set it to ``None`` or any other robot falsy value.
          Run on failure is not applied when library methods are executed directly from Python.
        - ``external_browser_executable`` <Dict <SupportedBrowsers, Path>>
          Dict mapping name of browser to path of executable of a browser.
          Will make opening new browsers of the given type use the set executablePath.
          Currently only configuring of `chromium` to a separate executable (chrome,
          chromium and Edge executables all work with recent versions) works.
        - ``jsextension`` <str>
          Path to Javascript module exposed as extra keywords. Module must be in CommonJS.
        - ``enable_presenter_mode`` <bool | dict>
          Automatic highlights to interacted components, slowMo and a small pause at the end. Can be enabled
          by giving True or can be customized by giving a dictionary: `{"duration": "2 seconds", "width": "2px",
          "style": "dotted", "color": "blue"}` Where `duration` is time format in Robot Framework format, defaults to
          2 seconds. `width` is width of the marker in pixels, defaults the `2px`. `style` is the style of border,
          defaults to `dotted`. `color` is the color of the marker, defaults to `blue`.
        - ``strict`` <bool>
          If keyword selector points multiple elements and keywords should interact with one element,
          keyword will fail if ``strict`` mode is true. Strict mode can be changed individually in keywords
          or by ```et Strict Mode`` keyword.
        """
        self.timeout = self.convert_timeout(timeout)
        self.retry_assertions_for = self.convert_timeout(retry_assertions_for)
        self.ROBOT_LIBRARY_LISTENER = self
        self._execution_stack: List[dict] = []
        self._running_on_failure_keyword = False
        self._pause_on_failure: Set["Browser"] = set()
        self.external_browser_executable: Dict[SupportedBrowsers, str] = (
            external_browser_executable or {}
        )
        self._unresolved_promises: Set[Future] = set()
        self._playwright_state = PlaywrightState(self)
        libraries = [
            self._playwright_state,
            Control(self),
            Cookie(self),
            Crawling(self),
            Devices(self),
            Evaluation(self),
            Formatter(self),
            Interaction(self),
            Getters(self),
            Network(self),
            RunOnFailureKeywords(self),
            StrictMode(self),
            Promises(self),
            Waiter(self),
            WebAppState(self),
        ]
        self.playwright = Playwright(
            self, enable_playwright_debug, playwright_process_port
        )
        self._auto_closing_level = auto_closing_level
        self.current_arguments = ()
        if jsextension is not None:
            libraries.append(self._initialize_jsextension(jsextension))
        self.presenter_mode = enable_presenter_mode
        self.strict_mode = strict
        self._keyword_formatters: dict = {}
        DynamicCore.__init__(self, libraries)
        # Parsing needs keywords to be discovered.
        self.run_on_failure_keyword = self._parse_run_on_failure_keyword(run_on_failure)

    def _parse_run_on_failure_keyword(
        self, keyword_name: Union[str, None]
    ) -> DelayedKeyword:
        if keyword_name is None or is_falsy(keyword_name):
            return DelayedKeyword(None, None, tuple(), {})
        parts = keyword_name.split("  ")
        keyword_name = parts[0]
        normalized_keyword_name = get_normalized_keyword(keyword_name)
        args = parts[1:]
        if normalized_keyword_name not in self.keywords:
            return DelayedKeyword(keyword_name, keyword_name, tuple(args), {})
        spec = PythonArgumentParser().parse(self.keywords[normalized_keyword_name])
        varargs = []
        kwargs = {}
        for arg in spec.resolve(args):
            for item in arg:
                if isinstance(item, tuple):
                    kwargs[item[0]] = item[1]
                else:
                    varargs.append(item)
        return DelayedKeyword(
            normalized_keyword_name, keyword_name, tuple(varargs), kwargs
        )

    def _initialize_jsextension(self, jsextension: str) -> LibraryComponent:
        component = LibraryComponent(self)
        with self.playwright.grpc_channel() as stub:
            response = stub.InitializeExtension(
                Request().FilePath(path=os.path.abspath(jsextension))
            )
            for name, args, doc in zip(
                response.keywords,
                response.keywordArguments,
                response.keywordDocumentations,
            ):
                self._jskeyword_call(component, name, args, doc)
        return component

    def _js_value_to_python_value(self, value: str) -> str:
        return {
            "true": "True",
            "false": "False",
            "null": "None",
            "undefined": "None",
            "NaN": "float('nan')",
            "Infinity": "float('inf')",
            "-Infinity": "float('-inf')",
        }.get(value, value)

    def _jskeyword_call(
        self,
        component: LibraryComponent,
        name: str,
        argument_names_and_default_values: str,
        doc: str,
    ):
        argument_names_and_vals = [
            [a.strip() for a in arg.split("=")]
            for arg in (argument_names_and_default_values or "").split(",")
            if arg
        ]
        argument_names_and_default_values_texts = []
        arg_set_texts = []
        for item in argument_names_and_vals:
            arg_name = item[0]
            if arg_name in ["logger", "playwright", "page"]:
                arg_set_texts.append(f'("{arg_name}", "RESERVED")')
            else:
                arg_set_texts.append(f'("{arg_name}", {arg_name})')
                if item[0] == "args":
                    argument_names_and_default_values_texts.append("*args")
                elif len(item) > 1:
                    argument_names_and_default_values_texts.append(
                        f"{arg_name}={self._js_value_to_python_value(item[1])}"
                    )
                else:
                    argument_names_and_default_values_texts.append(f"{arg_name}")
        text = f"""
@keyword
def {name}(self, {", ".join(argument_names_and_default_values_texts)}):
    \"\"\"{doc}\"\"\"
    _args_browser_internal = dict()
    _args_browser_internal["arguments"] = [{", ".join(arg_set_texts)}]
    with self.playwright.grpc_channel() as stub:
        responses = stub.CallExtensionKeyword(
            Request().KeywordCall(name="{name}", arguments=json.dumps(_args_browser_internal))
        )
        for response in responses:
            logger.info(response.log)
        if response.json == "":
            return
        return json.loads(response.json)
"""
        try:
            exec(
                text,
                {**globals(), "keyword": keyword, "json": json},
                component.__dict__,
            )
            setattr(
                component, name, types.MethodType(component.__dict__[name], component)
            )
        except SyntaxError as e:
            raise DataError(f"{e.msg} in {name}")

    @property
    def outputdir(self) -> str:
        if EXECUTION_CONTEXTS.current:
            return BuiltIn().get_variable_value("${OUTPUTDIR}")
        else:
            return "."

    @property
    def browser_output(self) -> Path:
        return Path(self.outputdir, "browser")

    def _start_suite(self, suite, result):
        if not self._suite_cleanup_done and self.browser_output.is_dir():
            self._suite_cleanup_done = True
            logger.debug(f"Removing: {self.browser_output}")
            shutil.rmtree(str(self.browser_output), ignore_errors=True)
        if self._auto_closing_level != AutoClosingLevel.MANUAL:
            try:
                self._execution_stack.append(self.get_browser_catalog())
            except ConnectionError as e:
                logger.debug(f"Browser._start_suite connection problem: {e}")

    def _start_test(self, test, result):
        if self._auto_closing_level == AutoClosingLevel.TEST:
            try:
                self._execution_stack.append(self.get_browser_catalog())
            except ConnectionError as e:
                logger.debug(f"Browser._start_test connection problem: {e}")

    def _end_test(self, test: TestCaseRunning, result: TestCaseResult):
        if len(self._unresolved_promises) > 0:
            logger.warn(f"Waiting unresolved promises at the end of test '{test.name}'")
            self.wait_for_all_promises()
        if self._auto_closing_level == AutoClosingLevel.TEST:
            if self.presenter_mode:
                logger.debug("Presenter mode: Wait for 5 seconds before pruning pages")
                time.sleep(5.0)
            if len(self._execution_stack) == 0:
                logger.debug("Browser._end_test empty execution stack")
                return
            try:
                catalog_before_test = self._execution_stack.pop()
                self._prune_execution_stack(catalog_before_test)
            except AssertionError as e:
                logger.debug(f"Test Case: {test.name}, End Test: {e}")
            except ConnectionError as e:
                logger.debug(f"Browser._end_test connection problem: {e}")

    def _end_suite(self, suite, result):
        if self._auto_closing_level != AutoClosingLevel.MANUAL:
            if len(self._execution_stack) == 0:
                logger.debug("Browser._end_suite empty execution stack")
                return
            try:
                catalog_before_suite = self._execution_stack.pop()
                self._prune_execution_stack(catalog_before_suite)
            except AssertionError as e:
                logger.debug(f"Test Suite: {suite.name}, End Suite: {e}")
            except ConnectionError as e:
                logger.debug(f"Browser._end_suite connection problem: {e}")

    def _prune_execution_stack(self, catalog_before: dict) -> None:
        catalog_after = self.get_browser_catalog()
        ctx_before_ids = [c["id"] for b in catalog_before for c in b["contexts"]]
        ctx_after_ids = [c["id"] for b in catalog_after for c in b["contexts"]]
        new_ctx_ids = [c for c in ctx_after_ids if c not in ctx_before_ids]
        for ctx_id in new_ctx_ids:
            self._playwright_state.switch_context(ctx_id)
            self._playwright_state.close_context()
        pages_before = [
            (p["id"], c["id"])
            for b in catalog_before
            for c in b["contexts"]
            for p in c["pages"]
        ]
        pages_after = [
            (p["id"], c["id"])
            for b in catalog_after
            for c in b["contexts"]
            for p in c["pages"]
            if c["id"] not in new_ctx_ids
        ]
        new_page_ids = [p for p in pages_after if p not in pages_before]
        for page_id, ctx_id in new_page_ids:
            self._playwright_state.close_page(page_id, ctx_id)

    def run_keyword(self, name, args, kwargs=None):
        try:
            return DynamicCore.run_keyword(self, name, args, kwargs)
        except AssertionError as e:
            self.keyword_error()
            if self._pause_on_failure:
                sys.__stdout__.write(f"\n[ FAIL ] {e}")
                sys.__stdout__.write(
                    "\n[Paused on failure] Press Enter to continue..\n"
                )
                sys.__stdout__.flush()
                input()
            raise e

    def start_keyword(self, name, attrs):
        """Take screenshot of tests that have failed due to timeout.

        This method is part of the Listener API implemented by the library.

        This can be done with BuiltIn keyword `Run Keyword If Timeout
        Occurred`, but the problem there is that you have to remember to
        put it into your Suite/Test Teardown. Since taking screenshot is
        the most obvious thing to do on failure, let's do it automatically.

        This cannot be implemented as a `end_test` listener method, since at
        that time, the teardown has already been executed and browser may have
        been closed already. This implementation will take the screenshot
        before the teardown begins to execute.
        """
        self.current_arguments = tuple(attrs["args"])
        if attrs["type"] == "Teardown":
            timeout_pattern = "Test timeout .* exceeded."
            test = EXECUTION_CONTEXTS.current.test
            if (
                test is not None
                and test.status == "FAIL"
                and re.match(timeout_pattern, test.message)
            ):
                self.screenshot_on_failure(test.name)

    def keyword_error(self):
        """Runs keyword on failure."""
        if self._running_on_failure_keyword or not self.run_on_failure_keyword.name:
            return
        self._running_on_failure_keyword = True
        varargs = self.run_on_failure_keyword.args
        kwargs = self.run_on_failure_keyword.kwargs
        try:
            if self.run_on_failure_keyword.name in self.keywords:
                if (
                    self.run_on_failure_keyword.name == "take_screenshot"
                    and not varargs
                ):
                    varargs = [self._failure_screenshot_path()]
                self.keywords[self.run_on_failure_keyword.name](*varargs, **kwargs)
            else:
                BuiltIn().run_keyword(
                    self.run_on_failure_keyword.name, *varargs, **kwargs
                )
        except Exception as err:
            logger.warn(
                f"Keyword '{self.run_on_failure_keyword}' could not be run on failure:\n{err}"
            )
        finally:
            self._running_on_failure_keyword = False

    def _failure_screenshot_path(self):
        valid_chars = "-_.() %s%s" % (string.ascii_letters, string.digits)
        test_name = BuiltIn().get_variable_value("${TEST NAME}", "GENERIC")
        return os.path.join(
            self.outputdir,
            "".join(c for c in test_name if c in valid_chars).replace(" ", "_")
            + "_FAILURE_SCREENSHOT_{index}",
        )

    def get_timeout(self, timeout: Union[timedelta, None]) -> float:
        if timeout is None:
            return self.timeout
        return self.convert_timeout(timeout)

    def convert_timeout(
        self, timeout: Union[timedelta, float], to_ms: bool = True
    ) -> float:
        convert = 1000 if to_ms else 1
        if isinstance(timeout, timedelta):
            return timeout.total_seconds() * convert
        return timestr_to_secs(timeout) * convert

    def millisecs_to_timestr(self, timeout: float) -> str:
        return secs_to_timestr(timeout / 1000)

    @overrides
    def get_keyword_documentation(self, name):
        doc = DynamicCore.get_keyword_documentation(self, name)
        if name == "__intro__":
            doc = doc.replace("%ASSERTION_TABLE%", AssertionOperator.__doc__)
            doc = doc.replace("%AUTO_CLOSING_LEVEL%", AutoClosingLevel.__doc__)
        elif name == "set_assertion_formatters":
            doc = doc.replace('"Keyword Name"', '"Get Text"')
            doc = f"{doc}\n    | ${{value}} =    `Get Text`    //div    ==    ${{SPACE}}Expected${{SPACE * 2}}Text"
            doc = f"{doc}\n    | Should Be Equal    ${{value}}    Expected Text\n"
        return doc



# Copyright 2020-     Robot Framework Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
from datetime import timedelta
from typing import Any, Dict, Optional

from typing_extensions import Literal

from ..base import LibraryComponent
from ..generated.playwright_pb2 import Request
from ..utils import keyword, logger
from ..utils.data_types import PageLoadStates, RequestMethod


def _get_headers(body: str, headers: Dict):
    try:
        json.loads(body)
        return {"Content-Type": "application/json", **headers}
    except json.decoder.JSONDecodeError:
        return headers


def _format_response(response: Dict):
    _jsonize_content(response, "body")
    if "request" in response:
        request = response["request"]
        _jsonize_content(request, "postData")
    logger.info(response)
    return response


def _jsonize_content(data, bodykey):
    headers = json.loads(data["headers"])
    data["headers"] = headers
    lower_headers = dict((k.lower(), v) for k, v in headers.items())
    if (
        "content-type" in lower_headers
        and "application/json" in lower_headers["content-type"]
    ):
        try:
            data[bodykey] = json.loads(data[bodykey])
        except json.decoder.JSONDecodeError:
            pass


class Network(LibraryComponent):
    @keyword(tags=("HTTP",))
    def http(
        self,
        url: str,
        method: RequestMethod = RequestMethod.GET,
        body: Optional[str] = None,
        headers: Optional[dict] = None,
    ) -> Any:
        """Performs an HTTP request in the current browser context

        Accepts the following arguments:
          - ``url`` The request url, e.g. ``/api/foo``.
          - ``method`` The HTTP method for the request. Defaults to GET.
          - ``body`` The request body. GET requests cannot have a body. If the body can be parsed as JSON,
          the ``Content-Type`` header for the request will be automatically set to ``application/json``.
          Defaults to None.
          - ``headers`` A dictionary of additional request headers. Defaults to None.

        The response is a Python dictionary with following attributes:
          - ``status`` <int> The status code of the response.
          - ``statusText`` <str> Status text corresponding to ``status``, e.g OK or INTERNAL SERVER ERROR.
          - ``body`` <dict> | <str> The response body. If the body can be parsed as a JSON obejct,
          it will be returned as Python dictionary, otherwise it is returned as a string.
          - ``headers`` <dict> A dictionary containing all response headers.
          - ``ok`` <bool> Whether the request was successfull, i.e. the ``status`` is range 200-299.

        Here's an example of using Robot Framework dictionary variables and extended variable syntax to
        do assertions on the response object:

        | &{res}=             `HTTP`                       /api/endpoint
        | Should Be Equal     ${res.status}              200
        | Should Be Equal     ${res.body.some_field}     some value

        """
        if headers is None:
            headers = {}
        body = body if body else ""
        with self.playwright.grpc_channel() as stub:
            response = stub.HttpRequest(
                Request().HttpRequest(
                    url=url,
                    method=method.name if method else "GET",
                    body=body,
                    headers=json.dumps(_get_headers(body, headers)),
                )
            )
            logger.debug(response.log)
            return _format_response(json.loads(response.json))

    def _wait_for_http(self, method: Literal["Request", "Response"], matcher, timeout):
        with self.playwright.grpc_channel() as stub:
            function = getattr(stub, f"WaitFor{method}")
            response = function(
                Request().HttpCapture(
                    urlOrPredicate=matcher,
                    timeout=self.get_timeout(timeout),
                )
            )
            logger.debug(response.log)
            # Add format response back here
            return (
                response.body
                if method == "Request"
                else _format_response(json.loads(response.json))
            )

    @keyword(tags=("Wait", "HTTP"))
    def wait_for_request(
        self, matcher: str = "", timeout: Optional[timedelta] = None
    ) -> Any:
        """Waits for request matching matcher to be made.

        ``matcher`` Request URL string, JavaScript regex or JavaScript function to match request by.
        By default (with empty string) matches first available request. For additional information, see the Playwright [https://playwright.dev/docs/api/class-page/#page-wait-for-request|waitForRequest documentation].

        ``timeout`` Timeout supports Robot Framework time format. Uses default timeout if not set.

        Example:
        | `Click`               \\#delayed_request
        | `Wait For Request`    timeout=1s

        Async Example:
        | ${promise} =    `Promise To`         `Wait For Request`    matcher=\\\\/\\\\/local\\\\w+\\\\:\\\\d+\\\\/api    timeout=3s
        | `Click`         \\#delayed_request
        | `Wait For`      ${promise}

        JavaScript Function Example:
        | `Click`               \\#delayed_request    # Creates response which should be waited before pressing save.
        | `Wait For Request`    [https://playwright.dev/docs/api/class-request|request] => request.url().endsWith('api/get/json') && request.method() === 'GET'
        """
        return self._wait_for_http("Request", matcher, timeout)

    @keyword(tags=("Wait", "HTTP"))
    def wait_for_response(
        self, matcher: str = "", timeout: Optional[timedelta] = None
    ) -> Any:
        """Waits for response matching matcher and returns python dict with contents.

        ``matcher`` Request URL string, JavaScript regex or JavaScript function to match request by.
        By default (with empty string) matches first available request. For additional information, see the Playwright [https://playwright.dev/docs/api/class-page/#page-wait-for-response|waitForResponse documentation].

        ``timeout`` Timeout in seconds. Uses default timeout if not set.

        The response is a Python dictionary with following attributes:
          - ``status`` <int> The status code of the response.
          - ``statusText`` <str> Status text corresponding to ``status``, e.g OK or INTERNAL SERVER ERROR.
          - ``body`` <dict | str> The response body. If the body can be parsed as a JSON obejct,
          it will be returned as Python dictionary, otherwise it is returned as a string.
          - ``headers`` <dict> A dictionary containing all response headers.
          - ``ok`` <bool> Whether the request was successfull, i.e. the ``status`` is range 200-299.
          - ``request`` <dict> containing ``method`` <str>, ``headers`` <dict> and ``postData`` <dict> | <str>
          - ``url`` <str> url of the request.

        Synchronous Example:
        | `Click`                \\#delayed_request    # Creates response which should be waited before next actions
        | `Wait For Response`    matcher=\\\\/\\\\/local\\\\w+\\\\:\\\\d+\\\\/api
        | `Click`                \\#save

        Asynchronous Example:
        | ${promise} =    `Promise To`    `Wait For Response`    timeout=60s
        | `Click`           \\#delayed_request    # Creates response which should be waited before pressing save.
        | `Click`           \\#next
        | `Wait For`        ${promise}            # Waits for the response
        | `Click`           \\#save

        JavaScript Function Example:
        | `Click`               \\#delayed_request    # Creates response which should be waited before pressing save.
        | `Wait For Response`   [https://playwright.dev/docs/api/class-response/|response] => response.url().endsWith('json') && response.request().method() === 'GET'
        """
        return self._wait_for_http("Response", matcher, timeout)

    @keyword(tags=("Wait", "HTTP"))
    def wait_until_network_is_idle(self, timeout: Optional[timedelta] = None):
        """Waits until there has been at least one instance of 500 ms of no network traffic on the page after loading.

        Doesn't wait for network traffic that wasn't initiated within 500ms of page load.

        ``timeout`` Timeout supports Robot Framework time format. Uses default timeout of 10 seconds if not set.

        Example:
        | `Go To`                         ${URL}
        | `Wait Until Network Is Idle`    timeout=3s
        """
        with self.playwright.grpc_channel() as stub:
            response = stub.WaitUntilNetworkIsIdle(
                Request().Timeout(timeout=self.get_timeout(timeout))
            )
            logger.debug(response.log)

    @keyword(tags=("Wait", "HTTP"))
    def wait_for_navigation(
        self,
        url: str,
        timeout: Optional[timedelta] = None,
        wait_until: PageLoadStates = PageLoadStates.load,
    ):
        """Waits until page has navigated to given ``url``.

        ``url``  expected navigation target address either the exact match or a JavaScript-like regex wrapped
         in ``/`` symbols.

        ``timeout`` Timeout in milliseconds. Uses default timeout of 10 seconds if not set.

        ``wait_until`` <"load"|"domcontentloaded"|"networkidle|"commit""> When to consider operation succeeded,
        defaults to load. Events can be either:
        'domcontentloaded' - consider operation to be finished when the DOMContentLoaded event is fired.
        'load' - consider operation to be finished when the load event is fired.
        'networkidle' - consider operation to be finished when there are no network connections for at least 500 ms.
        `commit` - consider operation to be finished when network response is received and the document started loading.

        Keyword works only when page is loaded and does not work if URL fragment changes. Example if
        https://marketsquare.github.io/robotframework-browser/Browser.html changes to
        https://marketsquare.github.io/robotframework-browser/Browser.html#Wait%20For%20Navigation
        keyword will fail.

        Example:
        | `Go To`                  ${ROOT_URL}/redirector.html
        | `Wait for navigation`    ${ROOT_URL}/posted.html    wait_until=${wait_until}
        """
        with self.playwright.grpc_channel() as stub:
            response = stub.WaitForNavigation(
                Request().UrlOptions(
                    url=Request().Url(
                        url=url, defaultTimeout=int(self.get_timeout(timeout))
                    ),
                    waitUntil=wait_until.name,
                )
            )
            logger.debug(response.log)



import urllib.parse
from typing import List, Optional, Set, Tuple

from robot.libraries.BuiltIn import BuiltIn  # type: ignore
from robotlibcore import keyword  # type: ignore

from Browser.base import LibraryComponent

from ..utils import logger


class Crawling(LibraryComponent):
    @keyword(tags=["Crawling"])
    def crawl_site(
        self,
        url: Optional[str] = None,
        page_crawl_keyword="take_screenshot",
        max_number_of_page_to_crawl: int = 1000,
        max_depth_to_crawl: int = 50,
    ):
        """
        Web crawler is a tool to go through all the pages on a specific URL domain.
        This happens by finding all links going to the same site and opening those.

        returns list of crawled urls.

        ``url`` is the page to start crawling from.

        ``page_crawl_keyword`` is the keyword that will be executed on every page.
        By default it will take a screenshot on every page.

        ``max_number_of_page_to_crawl`` is the upper limit of pages to crawl.
        Crawling will stop if the number of crawled pages goes over this.

        ``max_depth_to_crawl`` is the upper limit of consecutive links followed from the start page.
        Crawling will stop if there are no more links under this depth.
        """
        if url:
            self.library.new_page(url)
        return list(
            self._crawl(
                self.library.get_url() or "",
                page_crawl_keyword,
                max_number_of_page_to_crawl,
                max_depth_to_crawl,
            )
        )

    def _crawl(
        self,
        start_url: str,
        page_crawl_keyword: str,
        max_number_of_page_to_crawl: int,
        max_depth_to_crawl: int,
    ):
        hrefs_to_crawl: List[Tuple[str, int]] = [(start_url, 0)]
        url_parts = urllib.parse.urlparse(start_url)
        baseurl = url_parts.scheme + "://" + url_parts.netloc
        crawled: Set[str] = set()
        while hrefs_to_crawl and len(crawled) < max_number_of_page_to_crawl:
            href, depth = hrefs_to_crawl.pop()
            if not href.startswith(baseurl):
                continue
            if href in crawled:
                continue
            logger.info(f"Crawling url {href}")
            logger.console(
                f"{len(crawled) + 1} / {len(crawled) + 1 + len(hrefs_to_crawl)} : Crawling url {href}"
            )
            try:
                self.library.go_to(href)
            except Exception as e:
                logger.warn(f"Exception while crawling {href}: {e}")
                continue
            BuiltIn().run_keyword(page_crawl_keyword)
            child_hrefs = self._gather_links(depth)
            crawled.add(href)
            hrefs_to_crawl = self._build_urls_to_crawl(
                child_hrefs, hrefs_to_crawl, crawled, baseurl, max_depth_to_crawl
            )
        return crawled

    def _gather_links(self, parent_depth: int) -> List[Tuple[str, int]]:
        link_elements = self.library.get_elements("//a[@href]")
        links: Set[str] = set()
        depth = parent_depth + 1
        for link_element in link_elements:
            href, normal_link = self.library.execute_javascript(
                "(e) => [e.href, !e.download]", link_element
            )
            if normal_link:
                links.add(href)
        return [(c, depth) for c in links]

    def _build_urls_to_crawl(
        self,
        new_hrefs_to_crawl: List[Tuple[str, int]],
        old_hrefs_to_crawl: List[Tuple[str, int]],
        crawled: Set[str],
        baseurl: str,
        max_depth: int,
    ) -> List[Tuple[str, int]]:
        new_hrefs = []
        for href, depth in new_hrefs_to_crawl:
            if depth > max_depth:
                continue
            if href in [h[0] for h in old_hrefs_to_crawl]:
                continue
            if href in crawled:
                continue
            if not href.startswith(baseurl):
                continue
            logger.debug(f"Adding link to {href}")
            new_hrefs.append((href, depth))
        return new_hrefs + old_hrefs_to_crawl



# Copyright 2020-     Robot Framework Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time
from typing import Optional

import wrapt  # type: ignore

from .utils import logger


@wrapt.decorator
def with_assertion_polling(wrapped, instance, args, kwargs):
    start = time.time()
    timeout = instance.timeout / 1000
    retry_assertions_until = instance.retry_assertions_for / 1000
    retries_start: Optional[float] = None
    tries = 1
    try:
        logger.stash_this_thread()
        while True:
            try:
                return wrapped(*args, **kwargs)
            except AssertionError as e:
                if retries_start is None:
                    retries_start = time.time()
                elapsed = time.time() - start
                elapsed_retries = time.time() - retries_start
                if elapsed >= timeout or elapsed_retries >= retry_assertions_until:
                    raise e
                tries += 1
                if timeout - elapsed > 0.01:
                    time.sleep(0.01)
                logger.clear_thread_stash()
    finally:
        logger.flush_and_delete_thread_stash()
        now = time.time()
        logger.debug(
            f"""Assertion polling statistics:
First element asserted in: {(retries_start or now) - start} seconds
Total tries: {tries}
Elapsed time in retries {now - (retries_start or now)} seconds"""
        )



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: playwright.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x10playwright.proto\"\xdd\x11\n\x07Request\x1a\x07\n\x05\x45mpty\x1a\x16\n\x05Label\x12\r\n\x05label\x18\x01 \x01(\t\x1aT\n\x11ScreenshotOptions\x12\x10\n\x08selector\x18\x01 \x01(\t\x12\x0c\n\x04mask\x18\x02 \x01(\t\x12\x0f\n\x07options\x18\x03 \x01(\t\x12\x0e\n\x06strict\x18\x04 \x01(\x08\x1a.\n\x0bKeywordCall\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x11\n\targuments\x18\x02 \x01(\t\x1a\x18\n\x08\x46ilePath\x12\x0c\n\x04path\x18\x01 \x01(\t\x1a@\n\x0e\x46ileBySelector\x12\x0c\n\x04path\x18\x01 \x01(\t\x12\x10\n\x08selector\x18\x02 \x01(\t\x12\x0e\n\x06strict\x18\x03 \x01(\x08\x1a\x14\n\x04Json\x12\x0c\n\x04\x62ody\x18\x01 \x01(\t\x1a\x32\n\x12MouseButtonOptions\x12\x0e\n\x06\x61\x63tion\x18\x01 \x01(\t\x12\x0c\n\x04json\x18\x02 \x01(\t\x1a/\n\x10KeyboardKeypress\x12\x0e\n\x06\x61\x63tion\x18\x01 \x01(\t\x12\x0b\n\x03key\x18\x02 \x01(\t\x1a\x44\n\x14KeyboardInputOptions\x12\x0e\n\x06\x61\x63tion\x18\x01 \x01(\t\x12\r\n\x05input\x18\x02 \x01(\t\x12\r\n\x05\x64\x65lay\x18\x03 \x01(\x05\x1a.\n\x07\x42rowser\x12\x0f\n\x07\x62rowser\x18\x01 \x01(\t\x12\x12\n\nrawOptions\x18\x02 \x01(\t\x1a_\n\x07\x43ontext\x12\x12\n\nrawOptions\x18\x01 \x01(\t\x12\x15\n\rhideRfBrowser\x18\x02 \x01(\x08\x12\x16\n\x0e\x64\x65\x66\x61ultTimeout\x18\x03 \x01(\x05\x12\x11\n\ttraceFile\x18\x04 \x01(\t\x1a\x32\n\x0bPermissions\x12\x13\n\x0bpermissions\x18\x01 \x03(\t\x12\x0e\n\x06origin\x18\x02 \x01(\t\x1a*\n\x03Url\x12\x0b\n\x03url\x18\x01 \x01(\t\x12\x16\n\x0e\x64\x65\x66\x61ultTimeout\x18\x02 \x01(\x05\x1a:\n\nUrlOptions\x12\x19\n\x03url\x18\x01 \x01(\x0b\x32\x0c.Request.Url\x12\x11\n\twaitUntil\x18\x02 \x01(\t\x1a.\n\x0e\x43onnectBrowser\x12\x0f\n\x07\x62rowser\x18\x01 \x01(\t\x12\x0b\n\x03url\x18\x02 \x01(\t\x1a:\n\tTextInput\x12\r\n\x05input\x18\x01 \x01(\t\x12\x10\n\x08selector\x18\x02 \x01(\t\x12\x0c\n\x04type\x18\x03 \x01(\x08\x1a\x45\n\x0f\x45lementProperty\x12\x10\n\x08selector\x18\x01 \x01(\t\x12\x10\n\x08property\x18\x02 \x01(\t\x12\x0e\n\x06strict\x18\x03 \x01(\x08\x1aX\n\x08TypeText\x12\x10\n\x08selector\x18\x01 \x01(\t\x12\x0c\n\x04text\x18\x02 \x01(\t\x12\r\n\x05\x64\x65lay\x18\x03 \x01(\x05\x12\r\n\x05\x63lear\x18\x04 \x01(\x08\x12\x0e\n\x06strict\x18\x05 \x01(\x08\x1aI\n\x08\x46illText\x12\x10\n\x08selector\x18\x01 \x01(\t\x12\x0c\n\x04text\x18\x02 \x01(\t\x12\x0e\n\x06strict\x18\x03 \x01(\x08\x12\r\n\x05\x66orce\x18\x04 \x01(\x08\x1a-\n\tClearText\x12\x10\n\x08selector\x18\x01 \x01(\t\x12\x0e\n\x06strict\x18\x02 \x01(\x08\x1a:\n\tPressKeys\x12\x10\n\x08selector\x18\x01 \x01(\t\x12\x0e\n\x06strict\x18\x02 \x01(\x08\x12\x0b\n\x03key\x18\x03 \x03(\t\x1a\x42\n\x0f\x45lementSelector\x12\x10\n\x08selector\x18\x01 \x01(\t\x12\x0e\n\x06strict\x18\x02 \x01(\x08\x12\r\n\x05\x66orce\x18\x03 \x01(\x08\x1a\x1a\n\x07Timeout\x12\x0f\n\x07timeout\x18\x01 \x01(\x02\x1a\x16\n\x05Index\x12\r\n\x05index\x18\x01 \x01(\t\x1a,\n\rIdWithTimeout\x12\n\n\x02id\x18\x01 \x01(\t\x12\x0f\n\x07timeout\x18\x02 \x01(\x02\x1a\x1b\n\x08StyleTag\x12\x0f\n\x07\x63ontent\x18\x01 \x01(\t\x1aO\n\x1a\x45lementSelectorWithOptions\x12\x10\n\x08selector\x18\x01 \x01(\t\x12\x0f\n\x07options\x18\x02 \x01(\t\x12\x0e\n\x06strict\x18\x03 \x01(\x08\x1a~\n\x1b\x45lementSelectorWithDuration\x12\x10\n\x08selector\x18\x01 \x01(\t\x12\x10\n\x08\x64uration\x18\x02 \x01(\x05\x12\r\n\x05width\x18\x03 \x01(\t\x12\r\n\x05style\x18\x04 \x01(\t\x12\r\n\x05\x63olor\x18\x05 \x01(\t\x12\x0e\n\x06strict\x18\x06 \x01(\x08\x1aN\n\x15SelectElementSelector\x12\x10\n\x08selector\x18\x01 \x01(\t\x12\x13\n\x0bmatcherJson\x18\x02 \x01(\t\x12\x0e\n\x06strict\x18\x03 \x01(\x08\x1a\x42\n\x0eJavascriptCode\x12\x0e\n\x06script\x18\x01 \x01(\t\x12\x10\n\x08selector\x18\x02 \x01(\t\x12\x0e\n\x06strict\x18\x03 \x01(\x08\x1a[\n\x16WaitForFunctionOptions\x12\x0e\n\x06script\x18\x01 \x01(\t\x12\x10\n\x08selector\x18\x02 \x01(\t\x12\x0f\n\x07options\x18\x03 \x01(\t\x12\x0e\n\x06strict\x18\x04 \x01(\x08\x1a \n\x10PlaywrightObject\x12\x0c\n\x04info\x18\x01 \x01(\t\x1a)\n\x08Viewport\x12\r\n\x05width\x18\x01 \x01(\x05\x12\x0e\n\x06height\x18\x02 \x01(\x05\x1aI\n\x0bHttpRequest\x12\x0b\n\x03url\x18\x01 \x01(\t\x12\x0e\n\x06method\x18\x02 \x01(\t\x12\x0c\n\x04\x62ody\x18\x03 \x01(\t\x12\x0f\n\x07headers\x18\x04 \x01(\t\x1a\x36\n\x0bHttpCapture\x12\x16\n\x0eurlOrPredicate\x18\x01 \x01(\t\x12\x0f\n\x07timeout\x18\x02 \x01(\x02\x1a\x16\n\x06\x44\x65vice\x12\x0c\n\x04name\x18\x01 \x01(\t\x1a\x37\n\x0b\x41lertAction\x12\x13\n\x0b\x61lertAction\x18\x01 \x01(\t\x12\x13\n\x0bpromptInput\x18\x02 \x01(\t\x1a\x15\n\x04\x42ool\x12\r\n\x05value\x18\x01 \x01(\x08\x1a\x61\n\x0b\x45valuateAll\x12\x10\n\x08selector\x18\x01 \x01(\t\x12\x0e\n\x06script\x18\x02 \x01(\t\x12\x0b\n\x03\x61rg\x18\x03 \x01(\t\x12\x13\n\x0b\x61llElements\x18\x04 \x01(\x08\x12\x0e\n\x06strict\x18\x05 \x01(\x08\"F\n\x05Types\x1a=\n\x0bSelectEntry\x12\r\n\x05value\x18\x02 \x01(\t\x12\r\n\x05label\x18\x03 \x01(\t\x12\x10\n\x08selected\x18\x04 \x01(\x08\"\x8c\x05\n\x08Response\x1a\x14\n\x05\x45mpty\x12\x0b\n\x03log\x18\x01 \x01(\t\x1a#\n\x06String\x12\x0b\n\x03log\x18\x01 \x01(\t\x12\x0c\n\x04\x62ody\x18\x02 \x01(\t\x1a\x62\n\x08Keywords\x12\x0b\n\x03log\x18\x01 \x01(\t\x12\x10\n\x08keywords\x18\x02 \x03(\t\x12\x1d\n\x15keywordDocumentations\x18\x03 \x03(\t\x12\x18\n\x10keywordArguments\x18\x04 \x03(\t\x1a!\n\x04\x42ool\x12\x0b\n\x03log\x18\x01 \x01(\t\x12\x0c\n\x04\x62ody\x18\x02 \x01(\x08\x1a \n\x03Int\x12\x0b\n\x03log\x18\x01 \x01(\t\x12\x0c\n\x04\x62ody\x18\x02 \x01(\x05\x1a+\n\x06Select\x12!\n\x05\x65ntry\x18\x01 \x03(\x0b\x32\x12.Types.SelectEntry\x1a!\n\x04Json\x12\x0b\n\x03log\x18\x01 \x01(\t\x12\x0c\n\x04json\x18\x02 \x01(\t\x1a\x38\n\x19JavascriptExecutionResult\x12\x0b\n\x03log\x18\x01 \x01(\t\x12\x0e\n\x06result\x18\x02 \x01(\t\x1aY\n\x12NewContextResponse\x12\n\n\x02id\x18\x01 \x01(\t\x12\x0b\n\x03log\x18\x02 \x01(\t\x12\x16\n\x0e\x63ontextOptions\x18\x03 \x01(\t\x12\x12\n\nnewBrowser\x18\x04 \x01(\x08\x1a\x63\n\x0fNewPageResponse\x12\x0b\n\x03log\x18\x01 \x01(\t\x12\x0c\n\x04\x62ody\x18\x02 \x01(\t\x12\r\n\x05video\x18\x03 \x01(\t\x12\x12\n\nnewBrowser\x18\x04 \x01(\x08\x12\x12\n\nnewContext\x18\x05 \x01(\x08\x1aR\n\x12PageReportResponse\x12\x0b\n\x03log\x18\x01 \x01(\t\x12\x0e\n\x06\x65rrors\x18\x02 \x01(\t\x12\x0f\n\x07\x63onsole\x18\x03 \x01(\t\x12\x0e\n\x06pageId\x18\x04 \x01(\t2\xb6%\n\nPlaywright\x12+\n\tAddCookie\x12\r.Request.Json\x1a\x0f.Response.Empty\x12,\n\nGetCookies\x12\x0e.Request.Empty\x1a\x0e.Response.Json\x12\x33\n\x10\x44\x65leteAllCookies\x12\x0e.Request.Empty\x1a\x0f.Response.Empty\x12>\n\x0eTakeScreenshot\x12\x1a.Request.ScreenshotOptions\x1a\x10.Response.String\x12%\n\x04GoTo\x12\x0c.Request.Url\x1a\x0f.Response.Empty\x12)\n\x06GoBack\x12\x0e.Request.Empty\x1a\x0f.Response.Empty\x12,\n\tGoForward\x12\x0e.Request.Empty\x1a\x0f.Response.Empty\x12,\n\x08GetTitle\x12\x0e.Request.Empty\x1a\x10.Response.String\x12:\n\x0fGetElementCount\x12\x18.Request.ElementSelector\x1a\r.Response.Int\x12.\n\x08TypeText\x12\x11.Request.TypeText\x1a\x0f.Response.Empty\x12.\n\x08\x46illText\x12\x11.Request.FillText\x1a\x0f.Response.Empty\x12\x30\n\tClearText\x12\x12.Request.ClearText\x1a\x0f.Response.Empty\x12<\n\x0eGetDomProperty\x12\x18.Request.ElementProperty\x1a\x10.Response.String\x12\x35\n\x07GetText\x12\x18.Request.ElementSelector\x1a\x10.Response.String\x12\x41\n\x13GetElementAttribute\x12\x18.Request.ElementProperty\x1a\x10.Response.String\x12;\n\x0fGetBoolProperty\x12\x18.Request.ElementProperty\x1a\x0e.Response.Bool\x12\x31\n\x0fGetViewportSize\x12\x0e.Request.Empty\x1a\x0e.Response.Json\x12*\n\x06GetUrl\x12\x0e.Request.Empty\x1a\x10.Response.String\x12\x31\n\rGetPageSource\x12\x0e.Request.Empty\x1a\x10.Response.String\x12,\n\x05Press\x12\x12.Request.PressKeys\x1a\x0f.Response.Empty\x12>\n\x10GetSelectContent\x12\x18.Request.ElementSelector\x1a\x10.Response.Select\x12@\n\x0cSelectOption\x12\x1e.Request.SelectElementSelector\x1a\x10.Response.Select\x12;\n\x0e\x44\x65selectOption\x12\x18.Request.ElementSelector\x1a\x0f.Response.Empty\x12:\n\rCheckCheckbox\x12\x18.Request.ElementSelector\x1a\x0f.Response.Empty\x12<\n\x0fUncheckCheckbox\x12\x18.Request.ElementSelector\x1a\x0f.Response.Empty\x12*\n\x06Health\x12\x0e.Request.Empty\x1a\x10.Response.String\x12\x38\n\nGetElement\x12\x18.Request.ElementSelector\x1a\x10.Response.String\x12\x37\n\x0bGetElements\x12\x18.Request.ElementSelector\x1a\x0e.Response.Json\x12<\n\x10GetElementStates\x12\x18.Request.ElementSelector\x1a\x0e.Response.Json\x12/\n\nSetTimeout\x12\x10.Request.Timeout\x1a\x0f.Response.Empty\x12\x31\n\x0b\x41\x64\x64StyleTag\x12\x11.Request.StyleTag\x1a\x0f.Response.Empty\x12J\n\x11HighlightElements\x12$.Request.ElementSelectorWithDuration\x1a\x0f.Response.Empty\x12(\n\x08\x44ownload\x12\x0c.Request.Url\x1a\x0e.Response.Json\x12=\n\x05\x43lick\x12#.Request.ElementSelectorWithOptions\x1a\x0f.Response.Empty\x12=\n\x05Hover\x12#.Request.ElementSelectorWithOptions\x1a\x0f.Response.Empty\x12\x32\n\x05\x46ocus\x12\x18.Request.ElementSelector\x1a\x0f.Response.Empty\x12L\n\x14WaitForElementsState\x12#.Request.ElementSelectorWithOptions\x1a\x0f.Response.Empty\x12\x42\n\x0fWaitForFunction\x12\x1f.Request.WaitForFunctionOptions\x1a\x0e.Response.Json\x12Q\n\x11\x45xecuteJavascript\x12\x17.Request.JavascriptCode\x1a#.Response.JavascriptExecutionResult\x12O\n\x12\x45valuateJavascript\x12\x14.Request.EvaluateAll\x1a#.Response.JavascriptExecutionResult\x12\x45\n\x0eRecordSelector\x12\x0e.Request.Label\x1a#.Response.JavascriptExecutionResult\x12\x43\n\x0cGetPageState\x12\x0e.Request.Empty\x1a#.Response.JavascriptExecutionResult\x12\x35\n\x0fSetViewportSize\x12\x11.Request.Viewport\x1a\x0f.Response.Empty\x12\x34\n\x08GetStyle\x12\x18.Request.ElementSelector\x1a\x0e.Response.Json\x12:\n\x0eGetBoundingBox\x12\x18.Request.ElementSelector\x1a\x0e.Response.Json\x12\x33\n\x0bHttpRequest\x12\x14.Request.HttpRequest\x1a\x0e.Response.Json\x12\x38\n\x0eWaitForRequest\x12\x14.Request.HttpCapture\x1a\x10.Response.String\x12\x37\n\x0fWaitForResponse\x12\x14.Request.HttpCapture\x1a\x0e.Response.Json\x12\x34\n\x0fWaitForDownload\x12\x11.Request.FilePath\x1a\x0e.Response.Json\x12;\n\x16WaitUntilNetworkIsIdle\x12\x10.Request.Timeout\x1a\x0f.Response.Empty\x12\x39\n\x11WaitForNavigation\x12\x13.Request.UrlOptions\x1a\x0f.Response.Empty\x12\x30\n\x0eSetGeolocation\x12\r.Request.Json\x1a\x0f.Response.Empty\x12,\n\tGetDevice\x12\x0f.Request.Device\x1a\x0e.Response.Json\x12,\n\nGetDevices\x12\x0e.Request.Empty\x1a\x0e.Response.Json\x12\x34\n\x0bHandleAlert\x12\x14.Request.AlertAction\x1a\x0f.Response.Empty\x12\x36\n\x0cWaitForAlert\x12\x14.Request.AlertAction\x1a\x10.Response.String\x12;\n\x0bMouseButton\x12\x1b.Request.MouseButtonOptions\x1a\x0f.Response.Empty\x12+\n\tMouseMove\x12\r.Request.Json\x1a\x0f.Response.Empty\x12\x39\n\x0bKeyboardKey\x12\x19.Request.KeyboardKeypress\x1a\x0f.Response.Empty\x12?\n\rKeyboardInput\x12\x1d.Request.KeyboardInputOptions\x1a\x0f.Response.Empty\x12;\n\x10GetTableRowIndex\x12\x18.Request.ElementSelector\x1a\r.Response.Int\x12<\n\x11GetTableCellIndex\x12\x18.Request.ElementSelector\x1a\r.Response.Int\x12\x30\n\nUploadFile\x12\x11.Request.FilePath\x1a\x0f.Response.Empty\x12@\n\x14UploadFileBySelector\x12\x17.Request.FileBySelector\x1a\x0f.Response.Empty\x12<\n\x0fScrollToElement\x12\x18.Request.ElementSelector\x1a\x0f.Response.Empty\x12,\n\nSetOffline\x12\r.Request.Bool\x1a\x0f.Response.Empty\x12)\n\x06Reload\x12\x0e.Request.Empty\x1a\x0f.Response.Empty\x12\x36\n\nSwitchPage\x12\x16.Request.IdWithTimeout\x1a\x10.Response.String\x12\x31\n\rSwitchContext\x12\x0e.Request.Index\x1a\x10.Response.String\x12\x31\n\rSwitchBrowser\x12\x0e.Request.Index\x1a\x10.Response.String\x12\x32\n\x07NewPage\x12\x0c.Request.Url\x1a\x19.Response.NewPageResponse\x12<\n\nNewContext\x12\x10.Request.Context\x1a\x1c.Response.NewContextResponse\x12\x30\n\nNewBrowser\x12\x10.Request.Browser\x1a\x10.Response.String\x12=\n\x10\x43onnectToBrowser\x12\x17.Request.ConnectBrowser\x1a\x10.Response.String\x12\x30\n\x0c\x43loseBrowser\x12\x0e.Request.Empty\x1a\x10.Response.String\x12\x33\n\x10\x43loseAllBrowsers\x12\x0e.Request.Empty\x1a\x0f.Response.Empty\x12/\n\x0c\x43loseContext\x12\x0e.Request.Empty\x1a\x0f.Response.Empty\x12\x39\n\tClosePage\x12\x0e.Request.Empty\x1a\x1c.Response.PageReportResponse\x12\x33\n\x11GetBrowserCatalog\x12\x0e.Request.Empty\x1a\x0e.Response.Json\x12\x36\n\x10SaveStorageState\x12\x11.Request.FilePath\x1a\x0f.Response.Empty\x12\x39\n\x10GrantPermissions\x12\x14.Request.Permissions\x1a\x0f.Response.Empty\x12\x33\n\x10\x43learPermissions\x12\x0e.Request.Empty\x1a\x0f.Response.Empty\x12<\n\x13InitializeExtension\x12\x11.Request.FilePath\x1a\x12.Response.Keywords\x12>\n\x14\x43\x61llExtensionKeyword\x12\x14.Request.KeywordCall\x1a\x0e.Response.Json0\x01\x12-\n\tSetPeerId\x12\x0e.Request.Index\x1a\x10.Response.Stringb\x06proto3')



_REQUEST = DESCRIPTOR.message_types_by_name['Request']
_REQUEST_EMPTY = _REQUEST.nested_types_by_name['Empty']
_REQUEST_LABEL = _REQUEST.nested_types_by_name['Label']
_REQUEST_SCREENSHOTOPTIONS = _REQUEST.nested_types_by_name['ScreenshotOptions']
_REQUEST_KEYWORDCALL = _REQUEST.nested_types_by_name['KeywordCall']
_REQUEST_FILEPATH = _REQUEST.nested_types_by_name['FilePath']
_REQUEST_FILEBYSELECTOR = _REQUEST.nested_types_by_name['FileBySelector']
_REQUEST_JSON = _REQUEST.nested_types_by_name['Json']
_REQUEST_MOUSEBUTTONOPTIONS = _REQUEST.nested_types_by_name['MouseButtonOptions']
_REQUEST_KEYBOARDKEYPRESS = _REQUEST.nested_types_by_name['KeyboardKeypress']
_REQUEST_KEYBOARDINPUTOPTIONS = _REQUEST.nested_types_by_name['KeyboardInputOptions']
_REQUEST_BROWSER = _REQUEST.nested_types_by_name['Browser']
_REQUEST_CONTEXT = _REQUEST.nested_types_by_name['Context']
_REQUEST_PERMISSIONS = _REQUEST.nested_types_by_name['Permissions']
_REQUEST_URL = _REQUEST.nested_types_by_name['Url']
_REQUEST_URLOPTIONS = _REQUEST.nested_types_by_name['UrlOptions']
_REQUEST_CONNECTBROWSER = _REQUEST.nested_types_by_name['ConnectBrowser']
_REQUEST_TEXTINPUT = _REQUEST.nested_types_by_name['TextInput']
_REQUEST_ELEMENTPROPERTY = _REQUEST.nested_types_by_name['ElementProperty']
_REQUEST_TYPETEXT = _REQUEST.nested_types_by_name['TypeText']
_REQUEST_FILLTEXT = _REQUEST.nested_types_by_name['FillText']
_REQUEST_CLEARTEXT = _REQUEST.nested_types_by_name['ClearText']
_REQUEST_PRESSKEYS = _REQUEST.nested_types_by_name['PressKeys']
_REQUEST_ELEMENTSELECTOR = _REQUEST.nested_types_by_name['ElementSelector']
_REQUEST_TIMEOUT = _REQUEST.nested_types_by_name['Timeout']
_REQUEST_INDEX = _REQUEST.nested_types_by_name['Index']
_REQUEST_IDWITHTIMEOUT = _REQUEST.nested_types_by_name['IdWithTimeout']
_REQUEST_STYLETAG = _REQUEST.nested_types_by_name['StyleTag']
_REQUEST_ELEMENTSELECTORWITHOPTIONS = _REQUEST.nested_types_by_name['ElementSelectorWithOptions']
_REQUEST_ELEMENTSELECTORWITHDURATION = _REQUEST.nested_types_by_name['ElementSelectorWithDuration']
_REQUEST_SELECTELEMENTSELECTOR = _REQUEST.nested_types_by_name['SelectElementSelector']
_REQUEST_JAVASCRIPTCODE = _REQUEST.nested_types_by_name['JavascriptCode']
_REQUEST_WAITFORFUNCTIONOPTIONS = _REQUEST.nested_types_by_name['WaitForFunctionOptions']
_REQUEST_PLAYWRIGHTOBJECT = _REQUEST.nested_types_by_name['PlaywrightObject']
_REQUEST_VIEWPORT = _REQUEST.nested_types_by_name['Viewport']
_REQUEST_HTTPREQUEST = _REQUEST.nested_types_by_name['HttpRequest']
_REQUEST_HTTPCAPTURE = _REQUEST.nested_types_by_name['HttpCapture']
_REQUEST_DEVICE = _REQUEST.nested_types_by_name['Device']
_REQUEST_ALERTACTION = _REQUEST.nested_types_by_name['AlertAction']
_REQUEST_BOOL = _REQUEST.nested_types_by_name['Bool']
_REQUEST_EVALUATEALL = _REQUEST.nested_types_by_name['EvaluateAll']
_TYPES = DESCRIPTOR.message_types_by_name['Types']
_TYPES_SELECTENTRY = _TYPES.nested_types_by_name['SelectEntry']
_RESPONSE = DESCRIPTOR.message_types_by_name['Response']
_RESPONSE_EMPTY = _RESPONSE.nested_types_by_name['Empty']
_RESPONSE_STRING = _RESPONSE.nested_types_by_name['String']
_RESPONSE_KEYWORDS = _RESPONSE.nested_types_by_name['Keywords']
_RESPONSE_BOOL = _RESPONSE.nested_types_by_name['Bool']
_RESPONSE_INT = _RESPONSE.nested_types_by_name['Int']
_RESPONSE_SELECT = _RESPONSE.nested_types_by_name['Select']
_RESPONSE_JSON = _RESPONSE.nested_types_by_name['Json']
_RESPONSE_JAVASCRIPTEXECUTIONRESULT = _RESPONSE.nested_types_by_name['JavascriptExecutionResult']
_RESPONSE_NEWCONTEXTRESPONSE = _RESPONSE.nested_types_by_name['NewContextResponse']
_RESPONSE_NEWPAGERESPONSE = _RESPONSE.nested_types_by_name['NewPageResponse']
_RESPONSE_PAGEREPORTRESPONSE = _RESPONSE.nested_types_by_name['PageReportResponse']
Request = _reflection.GeneratedProtocolMessageType('Request', (_message.Message,), {

  'Empty' : _reflection.GeneratedProtocolMessageType('Empty', (_message.Message,), {
    'DESCRIPTOR' : _REQUEST_EMPTY,
    '__module__' : 'playwright_pb2'
    # @@protoc_insertion_point(class_scope:Request.Empty)
    })
  ,

  'Label' : _reflection.GeneratedProtocolMessageType('Label', (_message.Message,), {
    'DESCRIPTOR' : _REQUEST_LABEL,
    '__module__' : 'playwright_pb2'
    # @@protoc_insertion_point(class_scope:Request.Label)
    })
  ,

  'ScreenshotOptions' : _reflection.GeneratedProtocolMessageType('ScreenshotOptions', (_message.Message,), {
    'DESCRIPTOR' : _REQUEST_SCREENSHOTOPTIONS,
    '__module__' : 'playwright_pb2'
    # @@protoc_insertion_point(class_scope:Request.ScreenshotOptions)
    })
  ,

  'KeywordCall' : _reflection.GeneratedProtocolMessageType('KeywordCall', (_message.Message,), {
    'DESCRIPTOR' : _REQUEST_KEYWORDCALL,
    '__module__' : 'playwright_pb2'
    # @@protoc_insertion_point(class_scope:Request.KeywordCall)
    })
  ,

  'FilePath' : _reflection.GeneratedProtocolMessageType('FilePath', (_message.Message,), {
    'DESCRIPTOR' : _REQUEST_FILEPATH,
    '__module__' : 'playwright_pb2'
    # @@protoc_insertion_point(class_scope:Request.FilePath)
    })
  ,

  'FileBySelector' : _reflection.GeneratedProtocolMessageType('FileBySelector', (_message.Message,), {
    'DESCRIPTOR' : _REQUEST_FILEBYSELECTOR,
    '__module__' : 'playwright_pb2'
    # @@protoc_insertion_point(class_scope:Request.FileBySelector)
    })
  ,

  'Json' : _reflection.GeneratedProtocolMessageType('Json', (_message.Message,), {
    'DESCRIPTOR' : _REQUEST_JSON,
    '__module__' : 'playwright_pb2'
    # @@protoc_insertion_point(class_scope:Request.Json)
    })
  ,

  'MouseButtonOptions' : _reflection.GeneratedProtocolMessageType('MouseButtonOptions', (_message.Message,), {
    'DESCRIPTOR' : _REQUEST_MOUSEBUTTONOPTIONS,
    '__module__' : 'playwright_pb2'
    # @@protoc_insertion_point(class_scope:Request.MouseButtonOptions)
    })
  ,

  'KeyboardKeypress' : _reflection.GeneratedProtocolMessageType('KeyboardKeypress', (_message.Message,), {
    'DESCRIPTOR' : _REQUEST_KEYBOARDKEYPRESS,
    '__module__' : 'playwright_pb2'
    # @@protoc_insertion_point(class_scope:Request.KeyboardKeypress)
    })
  ,

  'KeyboardInputOptions' : _reflection.GeneratedProtocolMessageType('KeyboardInputOptions', (_message.Message,), {
    'DESCRIPTOR' : _REQUEST_KEYBOARDINPUTOPTIONS,
    '__module__' : 'playwright_pb2'
    # @@protoc_insertion_point(class_scope:Request.KeyboardInputOptions)
    })
  ,

  'Browser' : _reflection.GeneratedProtocolMessageType('Browser', (_message.Message,), {
    'DESCRIPTOR' : _REQUEST_BROWSER,
    '__module__' : 'playwright_pb2'
    # @@protoc_insertion_point(class_scope:Request.Browser)
    })
  ,

  'Context' : _reflection.GeneratedProtocolMessageType('Context', (_message.Message,), {
    'DESCRIPTOR' : _REQUEST_CONTEXT,
    '__module__' : 'playwright_pb2'
    # @@protoc_insertion_point(class_scope:Request.Context)
    })
  ,

  'Permissions' : _reflection.GeneratedProtocolMessageType('Permissions', (_message.Message,), {
    'DESCRIPTOR' : _REQUEST_PERMISSIONS,
    '__module__' : 'playwright_pb2'
    # @@protoc_insertion_point(class_scope:Request.Permissions)
    })
  ,

  'Url' : _reflection.GeneratedProtocolMessageType('Url', (_message.Message,), {
    'DESCRIPTOR' : _REQUEST_URL,
    '__module__' : 'playwright_pb2'
    # @@protoc_insertion_point(class_scope:Request.Url)
    })
  ,

  'UrlOptions' : _reflection.GeneratedProtocolMessageType('UrlOptions', (_message.Message,), {
    'DESCRIPTOR' : _REQUEST_URLOPTIONS,
    '__module__' : 'playwright_pb2'
    # @@protoc_insertion_point(class_scope:Request.UrlOptions)
    })
  ,

  'ConnectBrowser' : _reflection.GeneratedProtocolMessageType('ConnectBrowser', (_message.Message,), {
    'DESCRIPTOR' : _REQUEST_CONNECTBROWSER,
    '__module__' : 'playwright_pb2'
    # @@protoc_insertion_point(class_scope:Request.ConnectBrowser)
    })
  ,

  'TextInput' : _reflection.GeneratedProtocolMessageType('TextInput', (_message.Message,), {
    'DESCRIPTOR' : _REQUEST_TEXTINPUT,
    '__module__' : 'playwright_pb2'
    # @@protoc_insertion_point(class_scope:Request.TextInput)
    })
  ,

  'ElementProperty' : _reflection.GeneratedProtocolMessageType('ElementProperty', (_message.Message,), {
    'DESCRIPTOR' : _REQUEST_ELEMENTPROPERTY,
    '__module__' : 'playwright_pb2'
    # @@protoc_insertion_point(class_scope:Request.ElementProperty)
    })
  ,

  'TypeText' : _reflection.GeneratedProtocolMessageType('TypeText', (_message.Message,), {
    'DESCRIPTOR' : _REQUEST_TYPETEXT,
    '__module__' : 'playwright_pb2'
    # @@protoc_insertion_point(class_scope:Request.TypeText)
    })
  ,

  'FillText' : _reflection.GeneratedProtocolMessageType('FillText', (_message.Message,), {
    'DESCRIPTOR' : _REQUEST_FILLTEXT,
    '__module__' : 'playwright_pb2'
    # @@protoc_insertion_point(class_scope:Request.FillText)
    })
  ,

  'ClearText' : _reflection.GeneratedProtocolMessageType('ClearText', (_message.Message,), {
    'DESCRIPTOR' : _REQUEST_CLEARTEXT,
    '__module__' : 'playwright_pb2'
    # @@protoc_insertion_point(class_scope:Request.ClearText)
    })
  ,

  'PressKeys' : _reflection.GeneratedProtocolMessageType('PressKeys', (_message.Message,), {
    'DESCRIPTOR' : _REQUEST_PRESSKEYS,
    '__module__' : 'playwright_pb2'
    # @@protoc_insertion_point(class_scope:Request.PressKeys)
    })
  ,

  'ElementSelector' : _reflection.GeneratedProtocolMessageType('ElementSelector', (_message.Message,), {
    'DESCRIPTOR' : _REQUEST_ELEMENTSELECTOR,
    '__module__' : 'playwright_pb2'
    # @@protoc_insertion_point(class_scope:Request.ElementSelector)
    })
  ,

  'Timeout' : _reflection.GeneratedProtocolMessageType('Timeout', (_message.Message,), {
    'DESCRIPTOR' : _REQUEST_TIMEOUT,
    '__module__' : 'playwright_pb2'
    # @@protoc_insertion_point(class_scope:Request.Timeout)
    })
  ,

  'Index' : _reflection.GeneratedProtocolMessageType('Index', (_message.Message,), {
    'DESCRIPTOR' : _REQUEST_INDEX,
    '__module__' : 'playwright_pb2'
    # @@protoc_insertion_point(class_scope:Request.Index)
    })
  ,

  'IdWithTimeout' : _reflection.GeneratedProtocolMessageType('IdWithTimeout', (_message.Message,), {
    'DESCRIPTOR' : _REQUEST_IDWITHTIMEOUT,
    '__module__' : 'playwright_pb2'
    # @@protoc_insertion_point(class_scope:Request.IdWithTimeout)
    })
  ,

  'StyleTag' : _reflection.GeneratedProtocolMessageType('StyleTag', (_message.Message,), {
    'DESCRIPTOR' : _REQUEST_STYLETAG,
    '__module__' : 'playwright_pb2'
    # @@protoc_insertion_point(class_scope:Request.StyleTag)
    })
  ,

  'ElementSelectorWithOptions' : _reflection.GeneratedProtocolMessageType('ElementSelectorWithOptions', (_message.Message,), {
    'DESCRIPTOR' : _REQUEST_ELEMENTSELECTORWITHOPTIONS,
    '__module__' : 'playwright_pb2'
    # @@protoc_insertion_point(class_scope:Request.ElementSelectorWithOptions)
    })
  ,

  'ElementSelectorWithDuration' : _reflection.GeneratedProtocolMessageType('ElementSelectorWithDuration', (_message.Message,), {
    'DESCRIPTOR' : _REQUEST_ELEMENTSELECTORWITHDURATION,
    '__module__' : 'playwright_pb2'
    # @@protoc_insertion_point(class_scope:Request.ElementSelectorWithDuration)
    })
  ,

  'SelectElementSelector' : _reflection.GeneratedProtocolMessageType('SelectElementSelector', (_message.Message,), {
    'DESCRIPTOR' : _REQUEST_SELECTELEMENTSELECTOR,
    '__module__' : 'playwright_pb2'
    # @@protoc_insertion_point(class_scope:Request.SelectElementSelector)
    })
  ,

  'JavascriptCode' : _reflection.GeneratedProtocolMessageType('JavascriptCode', (_message.Message,), {
    'DESCRIPTOR' : _REQUEST_JAVASCRIPTCODE,
    '__module__' : 'playwright_pb2'
    # @@protoc_insertion_point(class_scope:Request.JavascriptCode)
    })
  ,

  'WaitForFunctionOptions' : _reflection.GeneratedProtocolMessageType('WaitForFunctionOptions', (_message.Message,), {
    'DESCRIPTOR' : _REQUEST_WAITFORFUNCTIONOPTIONS,
    '__module__' : 'playwright_pb2'
    # @@protoc_insertion_point(class_scope:Request.WaitForFunctionOptions)
    })
  ,

  'PlaywrightObject' : _reflection.GeneratedProtocolMessageType('PlaywrightObject', (_message.Message,), {
    'DESCRIPTOR' : _REQUEST_PLAYWRIGHTOBJECT,
    '__module__' : 'playwright_pb2'
    # @@protoc_insertion_point(class_scope:Request.PlaywrightObject)
    })
  ,

  'Viewport' : _reflection.GeneratedProtocolMessageType('Viewport', (_message.Message,), {
    'DESCRIPTOR' : _REQUEST_VIEWPORT,
    '__module__' : 'playwright_pb2'
    # @@protoc_insertion_point(class_scope:Request.Viewport)
    })
  ,

  'HttpRequest' : _reflection.GeneratedProtocolMessageType('HttpRequest', (_message.Message,), {
    'DESCRIPTOR' : _REQUEST_HTTPREQUEST,
    '__module__' : 'playwright_pb2'
    # @@protoc_insertion_point(class_scope:Request.HttpRequest)
    })
  ,

  'HttpCapture' : _reflection.GeneratedProtocolMessageType('HttpCapture', (_message.Message,), {
    'DESCRIPTOR' : _REQUEST_HTTPCAPTURE,
    '__module__' : 'playwright_pb2'
    # @@protoc_insertion_point(class_scope:Request.HttpCapture)
    })
  ,

  'Device' : _reflection.GeneratedProtocolMessageType('Device', (_message.Message,), {
    'DESCRIPTOR' : _REQUEST_DEVICE,
    '__module__' : 'playwright_pb2'
    # @@protoc_insertion_point(class_scope:Request.Device)
    })
  ,

  'AlertAction' : _reflection.GeneratedProtocolMessageType('AlertAction', (_message.Message,), {
    'DESCRIPTOR' : _REQUEST_ALERTACTION,
    '__module__' : 'playwright_pb2'
    # @@protoc_insertion_point(class_scope:Request.AlertAction)
    })
  ,

  'Bool' : _reflection.GeneratedProtocolMessageType('Bool', (_message.Message,), {
    'DESCRIPTOR' : _REQUEST_BOOL,
    '__module__' : 'playwright_pb2'
    # @@protoc_insertion_point(class_scope:Request.Bool)
    })
  ,

  'EvaluateAll' : _reflection.GeneratedProtocolMessageType('EvaluateAll', (_message.Message,), {
    'DESCRIPTOR' : _REQUEST_EVALUATEALL,
    '__module__' : 'playwright_pb2'
    # @@protoc_insertion_point(class_scope:Request.EvaluateAll)
    })
  ,
  'DESCRIPTOR' : _REQUEST,
  '__module__' : 'playwright_pb2'
  # @@protoc_insertion_point(class_scope:Request)
  })
_sym_db.RegisterMessage(Request)
_sym_db.RegisterMessage(Request.Empty)
_sym_db.RegisterMessage(Request.Label)
_sym_db.RegisterMessage(Request.ScreenshotOptions)
_sym_db.RegisterMessage(Request.KeywordCall)
_sym_db.RegisterMessage(Request.FilePath)
_sym_db.RegisterMessage(Request.FileBySelector)
_sym_db.RegisterMessage(Request.Json)
_sym_db.RegisterMessage(Request.MouseButtonOptions)
_sym_db.RegisterMessage(Request.KeyboardKeypress)
_sym_db.RegisterMessage(Request.KeyboardInputOptions)
_sym_db.RegisterMessage(Request.Browser)
_sym_db.RegisterMessage(Request.Context)
_sym_db.RegisterMessage(Request.Permissions)
_sym_db.RegisterMessage(Request.Url)
_sym_db.RegisterMessage(Request.UrlOptions)
_sym_db.RegisterMessage(Request.ConnectBrowser)
_sym_db.RegisterMessage(Request.TextInput)
_sym_db.RegisterMessage(Request.ElementProperty)
_sym_db.RegisterMessage(Request.TypeText)
_sym_db.RegisterMessage(Request.FillText)
_sym_db.RegisterMessage(Request.ClearText)
_sym_db.RegisterMessage(Request.PressKeys)
_sym_db.RegisterMessage(Request.ElementSelector)
_sym_db.RegisterMessage(Request.Timeout)
_sym_db.RegisterMessage(Request.Index)
_sym_db.RegisterMessage(Request.IdWithTimeout)
_sym_db.RegisterMessage(Request.StyleTag)
_sym_db.RegisterMessage(Request.ElementSelectorWithOptions)
_sym_db.RegisterMessage(Request.ElementSelectorWithDuration)
_sym_db.RegisterMessage(Request.SelectElementSelector)
_sym_db.RegisterMessage(Request.JavascriptCode)
_sym_db.RegisterMessage(Request.WaitForFunctionOptions)
_sym_db.RegisterMessage(Request.PlaywrightObject)
_sym_db.RegisterMessage(Request.Viewport)
_sym_db.RegisterMessage(Request.HttpRequest)
_sym_db.RegisterMessage(Request.HttpCapture)
_sym_db.RegisterMessage(Request.Device)
_sym_db.RegisterMessage(Request.AlertAction)
_sym_db.RegisterMessage(Request.Bool)
_sym_db.RegisterMessage(Request.EvaluateAll)

Types = _reflection.GeneratedProtocolMessageType('Types', (_message.Message,), {

  'SelectEntry' : _reflection.GeneratedProtocolMessageType('SelectEntry', (_message.Message,), {
    'DESCRIPTOR' : _TYPES_SELECTENTRY,
    '__module__' : 'playwright_pb2'
    # @@protoc_insertion_point(class_scope:Types.SelectEntry)
    })
  ,
  'DESCRIPTOR' : _TYPES,
  '__module__' : 'playwright_pb2'
  # @@protoc_insertion_point(class_scope:Types)
  })
_sym_db.RegisterMessage(Types)
_sym_db.RegisterMessage(Types.SelectEntry)

Response = _reflection.GeneratedProtocolMessageType('Response', (_message.Message,), {

  'Empty' : _reflection.GeneratedProtocolMessageType('Empty', (_message.Message,), {
    'DESCRIPTOR' : _RESPONSE_EMPTY,
    '__module__' : 'playwright_pb2'
    # @@protoc_insertion_point(class_scope:Response.Empty)
    })
  ,

  'String' : _reflection.GeneratedProtocolMessageType('String', (_message.Message,), {
    'DESCRIPTOR' : _RESPONSE_STRING,
    '__module__' : 'playwright_pb2'
    # @@protoc_insertion_point(class_scope:Response.String)
    })
  ,

  'Keywords' : _reflection.GeneratedProtocolMessageType('Keywords', (_message.Message,), {
    'DESCRIPTOR' : _RESPONSE_KEYWORDS,
    '__module__' : 'playwright_pb2'
    # @@protoc_insertion_point(class_scope:Response.Keywords)
    })
  ,

  'Bool' : _reflection.GeneratedProtocolMessageType('Bool', (_message.Message,), {
    'DESCRIPTOR' : _RESPONSE_BOOL,
    '__module__' : 'playwright_pb2'
    # @@protoc_insertion_point(class_scope:Response.Bool)
    })
  ,

  'Int' : _reflection.GeneratedProtocolMessageType('Int', (_message.Message,), {
    'DESCRIPTOR' : _RESPONSE_INT,
    '__module__' : 'playwright_pb2'
    # @@protoc_insertion_point(class_scope:Response.Int)
    })
  ,

  'Select' : _reflection.GeneratedProtocolMessageType('Select', (_message.Message,), {
    'DESCRIPTOR' : _RESPONSE_SELECT,
    '__module__' : 'playwright_pb2'
    # @@protoc_insertion_point(class_scope:Response.Select)
    })
  ,

  'Json' : _reflection.GeneratedProtocolMessageType('Json', (_message.Message,), {
    'DESCRIPTOR' : _RESPONSE_JSON,
    '__module__' : 'playwright_pb2'
    # @@protoc_insertion_point(class_scope:Response.Json)
    })
  ,

  'JavascriptExecutionResult' : _reflection.GeneratedProtocolMessageType('JavascriptExecutionResult', (_message.Message,), {
    'DESCRIPTOR' : _RESPONSE_JAVASCRIPTEXECUTIONRESULT,
    '__module__' : 'playwright_pb2'
    # @@protoc_insertion_point(class_scope:Response.JavascriptExecutionResult)
    })
  ,

  'NewContextResponse' : _reflection.GeneratedProtocolMessageType('NewContextResponse', (_message.Message,), {
    'DESCRIPTOR' : _RESPONSE_NEWCONTEXTRESPONSE,
    '__module__' : 'playwright_pb2'
    # @@protoc_insertion_point(class_scope:Response.NewContextResponse)
    })
  ,

  'NewPageResponse' : _reflection.GeneratedProtocolMessageType('NewPageResponse', (_message.Message,), {
    'DESCRIPTOR' : _RESPONSE_NEWPAGERESPONSE,
    '__module__' : 'playwright_pb2'
    # @@protoc_insertion_point(class_scope:Response.NewPageResponse)
    })
  ,

  'PageReportResponse' : _reflection.GeneratedProtocolMessageType('PageReportResponse', (_message.Message,), {
    'DESCRIPTOR' : _RESPONSE_PAGEREPORTRESPONSE,
    '__module__' : 'playwright_pb2'
    # @@protoc_insertion_point(class_scope:Response.PageReportResponse)
    })
  ,
  'DESCRIPTOR' : _RESPONSE,
  '__module__' : 'playwright_pb2'
  # @@protoc_insertion_point(class_scope:Response)
  })
_sym_db.RegisterMessage(Response)
_sym_db.RegisterMessage(Response.Empty)
_sym_db.RegisterMessage(Response.String)
_sym_db.RegisterMessage(Response.Keywords)
_sym_db.RegisterMessage(Response.Bool)
_sym_db.RegisterMessage(Response.Int)
_sym_db.RegisterMessage(Response.Select)
_sym_db.RegisterMessage(Response.Json)
_sym_db.RegisterMessage(Response.JavascriptExecutionResult)
_sym_db.RegisterMessage(Response.NewContextResponse)
_sym_db.RegisterMessage(Response.NewPageResponse)
_sym_db.RegisterMessage(Response.PageReportResponse)

_PLAYWRIGHT = DESCRIPTOR.services_by_name['Playwright']
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _REQUEST._serialized_start=21
  _REQUEST._serialized_end=2290
  _REQUEST_EMPTY._serialized_start=32
  _REQUEST_EMPTY._serialized_end=39
  _REQUEST_LABEL._serialized_start=41
  _REQUEST_LABEL._serialized_end=63
  _REQUEST_SCREENSHOTOPTIONS._serialized_start=65
  _REQUEST_SCREENSHOTOPTIONS._serialized_end=149
  _REQUEST_KEYWORDCALL._serialized_start=151
  _REQUEST_KEYWORDCALL._serialized_end=197
  _REQUEST_FILEPATH._serialized_start=199
  _REQUEST_FILEPATH._serialized_end=223
  _REQUEST_FILEBYSELECTOR._serialized_start=225
  _REQUEST_FILEBYSELECTOR._serialized_end=289
  _REQUEST_JSON._serialized_start=291
  _REQUEST_JSON._serialized_end=311
  _REQUEST_MOUSEBUTTONOPTIONS._serialized_start=313
  _REQUEST_MOUSEBUTTONOPTIONS._serialized_end=363
  _REQUEST_KEYBOARDKEYPRESS._serialized_start=365
  _REQUEST_KEYBOARDKEYPRESS._serialized_end=412
  _REQUEST_KEYBOARDINPUTOPTIONS._serialized_start=414
  _REQUEST_KEYBOARDINPUTOPTIONS._serialized_end=482
  _REQUEST_BROWSER._serialized_start=484
  _REQUEST_BROWSER._serialized_end=530
  _REQUEST_CONTEXT._serialized_start=532
  _REQUEST_CONTEXT._serialized_end=627
  _REQUEST_PERMISSIONS._serialized_start=629
  _REQUEST_PERMISSIONS._serialized_end=679
  _REQUEST_URL._serialized_start=681
  _REQUEST_URL._serialized_end=723
  _REQUEST_URLOPTIONS._serialized_start=725
  _REQUEST_URLOPTIONS._serialized_end=783
  _REQUEST_CONNECTBROWSER._serialized_start=785
  _REQUEST_CONNECTBROWSER._serialized_end=831
  _REQUEST_TEXTINPUT._serialized_start=833
  _REQUEST_TEXTINPUT._serialized_end=891
  _REQUEST_ELEMENTPROPERTY._serialized_start=893
  _REQUEST_ELEMENTPROPERTY._serialized_end=962
  _REQUEST_TYPETEXT._serialized_start=964
  _REQUEST_TYPETEXT._serialized_end=1052
  _REQUEST_FILLTEXT._serialized_start=1054
  _REQUEST_FILLTEXT._serialized_end=1127
  _REQUEST_CLEARTEXT._serialized_start=1129
  _REQUEST_CLEARTEXT._serialized_end=1174
  _REQUEST_PRESSKEYS._serialized_start=1176
  _REQUEST_PRESSKEYS._serialized_end=1234
  _REQUEST_ELEMENTSELECTOR._serialized_start=1236
  _REQUEST_ELEMENTSELECTOR._serialized_end=1302
  _REQUEST_TIMEOUT._serialized_start=1304
  _REQUEST_TIMEOUT._serialized_end=1330
  _REQUEST_INDEX._serialized_start=1332
  _REQUEST_INDEX._serialized_end=1354
  _REQUEST_IDWITHTIMEOUT._serialized_start=1356
  _REQUEST_IDWITHTIMEOUT._serialized_end=1400
  _REQUEST_STYLETAG._serialized_start=1402
  _REQUEST_STYLETAG._serialized_end=1429
  _REQUEST_ELEMENTSELECTORWITHOPTIONS._serialized_start=1431
  _REQUEST_ELEMENTSELECTORWITHOPTIONS._serialized_end=1510
  _REQUEST_ELEMENTSELECTORWITHDURATION._serialized_start=1512
  _REQUEST_ELEMENTSELECTORWITHDURATION._serialized_end=1638
  _REQUEST_SELECTELEMENTSELECTOR._serialized_start=1640
  _REQUEST_SELECTELEMENTSELECTOR._serialized_end=1718
  _REQUEST_JAVASCRIPTCODE._serialized_start=1720
  _REQUEST_JAVASCRIPTCODE._serialized_end=1786
  _REQUEST_WAITFORFUNCTIONOPTIONS._serialized_start=1788
  _REQUEST_WAITFORFUNCTIONOPTIONS._serialized_end=1879
  _REQUEST_PLAYWRIGHTOBJECT._serialized_start=1881
  _REQUEST_PLAYWRIGHTOBJECT._serialized_end=1913
  _REQUEST_VIEWPORT._serialized_start=1915
  _REQUEST_VIEWPORT._serialized_end=1956
  _REQUEST_HTTPREQUEST._serialized_start=1958
  _REQUEST_HTTPREQUEST._serialized_end=2031
  _REQUEST_HTTPCAPTURE._serialized_start=2033
  _REQUEST_HTTPCAPTURE._serialized_end=2087
  _REQUEST_DEVICE._serialized_start=2089
  _REQUEST_DEVICE._serialized_end=2111
  _REQUEST_ALERTACTION._serialized_start=2113
  _REQUEST_ALERTACTION._serialized_end=2168
  _REQUEST_BOOL._serialized_start=2170
  _REQUEST_BOOL._serialized_end=2191
  _REQUEST_EVALUATEALL._serialized_start=2193
  _REQUEST_EVALUATEALL._serialized_end=2290
  _TYPES._serialized_start=2292
  _TYPES._serialized_end=2362
  _TYPES_SELECTENTRY._serialized_start=2301
  _TYPES_SELECTENTRY._serialized_end=2362
  _RESPONSE._serialized_start=2365
  _RESPONSE._serialized_end=3017
  _RESPONSE_EMPTY._serialized_start=2377
  _RESPONSE_EMPTY._serialized_end=2397
  _RESPONSE_STRING._serialized_start=2399
  _RESPONSE_STRING._serialized_end=2434
  _RESPONSE_KEYWORDS._serialized_start=2436
  _RESPONSE_KEYWORDS._serialized_end=2534
  _RESPONSE_BOOL._serialized_start=2536
  _RESPONSE_BOOL._serialized_end=2569
  _RESPONSE_INT._serialized_start=2571
  _RESPONSE_INT._serialized_end=2603
  _RESPONSE_SELECT._serialized_start=2605
  _RESPONSE_SELECT._serialized_end=2648
  _RESPONSE_JSON._serialized_start=2650
  _RESPONSE_JSON._serialized_end=2683
  _RESPONSE_JAVASCRIPTEXECUTIONRESULT._serialized_start=2685
  _RESPONSE_JAVASCRIPTEXECUTIONRESULT._serialized_end=2741
  _RESPONSE_NEWCONTEXTRESPONSE._serialized_start=2743
  _RESPONSE_NEWCONTEXTRESPONSE._serialized_end=2832
  _RESPONSE_NEWPAGERESPONSE._serialized_start=2834
  _RESPONSE_NEWPAGERESPONSE._serialized_end=2933
  _RESPONSE_PAGEREPORTRESPONSE._serialized_start=2935
  _RESPONSE_PAGEREPORTRESPONSE._serialized_end=3017
  _PLAYWRIGHT._serialized_start=3020
  _PLAYWRIGHT._serialized_end=7810
# @@protoc_insertion_point(module_scope)



class ContextCache:
    def __init__(self):
        self.cache = {}

    def add(self, cache_uuid: str, item: dict):
        self.cache[cache_uuid] = item

    def remove(self, cache_uuid: str):
        self.cache.pop(cache_uuid, None)

    def get(self, cache_uuid: str):
        return self.cache.get(cache_uuid, None)



# Copyright 2020-     Robot Framework Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import time
from datetime import timedelta
from typing import Dict, Optional, Union

from ..base import LibraryComponent
from ..generated.playwright_pb2 import Request
from ..utils import ElementState, keyword, logger


class Waiter(LibraryComponent):
    @keyword(tags=("Wait", "PageContent"))
    def wait_for_elements_state(
        self,
        selector: str,
        state: ElementState = ElementState.visible,
        timeout: Optional[timedelta] = None,
        message: Optional[str] = None,
    ):
        """Waits for the element found by ``selector`` to satisfy state option.

        Note that Browser library has `Implicit waiting` mechanisms. Depending on
        the situation you might not need to use `Wait for Elements State`.

        State options could be either appear/disappear from dom, or become visible/hidden.
        If at the moment of calling the keyword, the selector already satisfies the condition,
        the keyword will return immediately.

        If the selector doesn't satisfy the condition within the timeout the keyword will FAIL.

        ``selector`` Selector of the corresponding object.
        See the `Finding elements` section for details about the selectors.

        ``state`` See `ElementState` for explanation.

        Note that element without any content or with display:none has an empty bounding box
        and is not considered visible.

        ``timeout`` uses default timeout from library if not set.

        ``message`` overrides the default error message. The ``message``
        argument accepts `{selector}`, `{function}`, and `{timeout}`
        [https://docs.python.org/3/library/stdtypes.html#str.format|format] options.
        The `{function}` formatter is same ``state`` argument value.

        Keyword uses strict mode, see `Finding elements` for more details about strict mode.

        Example:
        | `Wait For Elements State`    //h1    visible    timeout=2 s
        | `Wait For Elements State`    //hi    focused    1s
        """
        timeout_as_str = self.millisecs_to_timestr(self.get_timeout(timeout))
        funct = {
            ElementState.readonly: "e => e.readOnly",
            ElementState.selected: "e => e.selected",
            ElementState.deselected: "e => !e.selected",
            ElementState.focused: "e => document.activeElement === e",
            ElementState.defocused: "e => document.activeElement !== e",
            ElementState.checked: "e => e.checked",
            ElementState.unchecked: "e => !e.checked",
        }
        if state in [
            ElementState.attached,
            ElementState.detached,
            ElementState.visible,
            ElementState.hidden,
            ElementState.stable,
            ElementState.enabled,
            ElementState.disabled,
            ElementState.editable,
        ]:
            end = float(
                self.convert_timeout(timeout, False) if timeout else self.timeout / 1000
            )
            end += time.monotonic()
            while True:
                try:
                    return self._wait_for_elements_state(
                        selector, state, timeout, self.strict_mode
                    )
                except Exception as error:
                    if end > time.monotonic():
                        logger.debug(f"Suppress error: {error}")
                    else:
                        if message:
                            message = message.format(
                                selector=selector,
                                function=state.name,
                                timeout=timeout_as_str,
                            )
                            raise AssertionError(message)
                        raise
        else:
            self.wait_for_function(
                funct[state],
                selector=selector,
                timeout=timeout,
                message=message,
            )

    def _wait_for_elements_state(
        self,
        selector: str,
        state: ElementState = ElementState.visible,
        timeout: Optional[timedelta] = None,
        strict: bool = True,
    ):
        with self.playwright.grpc_channel() as stub:
            options: Dict[str, object] = {"state": state.name}
            if timeout:
                options["timeout"] = self.get_timeout(timeout)
            options_json = json.dumps(options)
            response = stub.WaitForElementsState(
                Request().ElementSelectorWithOptions(
                    selector=selector, options=options_json, strict=strict
                )
            )
            logger.info(response.log)

    @keyword(tags=("Wait", "PageContent"))
    def wait_for_function(
        self,
        function: str,
        selector: str = "",
        polling: Union[str, timedelta] = "raf",
        timeout: Optional[timedelta] = None,
        message: Optional[str] = None,
    ):
        """Polls JavaScript expression or function in browser until it returns a (JavaScript) truthy value.

        ``function`` A valid javascript function or a javascript function body. For example
        ``() => true`` and ``true`` will behave similarly.

        ``selector`` Selector to resolve and pass to the JavaScript function. This will be the first
        argument the function receives. If given a selector a function is necessary, with an argument
        to capture the elementhandle. For example ``(element) => document.activeElement === element``
        See the `Finding elements` section for details about the selectors.

        ``polling`` Default polling value of "raf" polls in a callback for ``requestAnimationFrame``.
        Any other value for polling will be parsed as a robot framework time for interval between polls.

        ``timeout`` Uses default timeout of 10 seconds if not set.

        ``message`` overrides the default error message. The ``message``
        argument accepts `{selector}`, `{function}`, and `{timeout}`
        [https://docs.python.org/3/library/stdtypes.html#str.format|format] options.

        Keyword uses strict mode, see `Finding elements` for more details about strict mode.

        Example usage:
        | ${promise}      `Promise To`      `Wait For Function`    element => element.style.width=="100%"    selector=\\#progress_bar    timeout=4s
        | `Click`         \\#progress_bar
        | `Wait For`      ${promise}
        """
        timeout_as_str = self.millisecs_to_timestr(self.get_timeout(timeout))
        end = float(
            self.convert_timeout(timeout, False) if timeout else self.timeout / 1000
        )
        end += time.monotonic()
        while True:
            try:
                return self._wait_for_function(
                    function, selector, polling, timeout, self.strict_mode
                )
            except Exception as error:
                if end > time.monotonic():
                    logger.debug(f"Suppress {error}")
                else:
                    if message:
                        message = message.format(
                            selector=selector, function=function, timeout=timeout_as_str
                        )
                        raise AssertionError(message)
                    raise

    def _wait_for_function(
        self,
        function: str,
        selector: str = "",
        polling: Union[str, timedelta] = "raf",
        timeout: Optional[timedelta] = None,
        strict: bool = True,
    ):
        with self.playwright.grpc_channel() as stub:
            options: Dict[str, int] = {}
            if polling != "raf":
                options["polling"] = self.convert_timeout(polling)  # type: ignore
            if timeout:
                options["timeout"] = self.convert_timeout(timeout)  # type: ignore
            options_json = json.dumps(options)
            response = stub.WaitForFunction(
                Request().WaitForFunctionOptions(
                    script=function,
                    selector=selector,
                    options=options_json,
                    strict=strict,
                )
            )
            logger.debug(response.json)
            logger.info(response.log)



# Copyright 2020-     Robot Framework Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import base64
import json
import os
from collections.abc import Iterable
from datetime import timedelta
from pathlib import Path
from typing import List, Optional, Union

from robot.utils import get_link_path  # type: ignore

from ..base import LibraryComponent
from ..generated.playwright_pb2 import Request
from ..utils import keyword, logger
from ..utils.data_types import BoundingBox, Permission, ScreenshotFileTypes


class Control(LibraryComponent):
    """Keywords to do things on the current browser page and modify the page"""

    @keyword(tags=("Setter", "BrowserControl"))
    def go_forward(self):
        """Navigates to the next page in history."""
        with self.playwright.grpc_channel() as stub:
            response = stub.GoForward(Request.Empty())
            logger.info(response.log)

    @keyword(tags=("Setter", "BrowserControl"))
    def go_back(self):
        """Navigates to the previous page in history."""
        with self.playwright.grpc_channel() as stub:
            response = stub.GoBack(Request.Empty())
            logger.info(response.log)

    @keyword(tags=("Setter", "BrowserControl"))
    def go_to(self, url: str, timeout: Optional[timedelta] = None):
        """Navigates to the given ``url``.

        ``url`` <str> URL to be navigated to.

        ``timeout`` <str> time to wait page to load. If not defined
        will use the library default timeout.
        """
        with self.playwright.grpc_channel() as stub:
            response = stub.GoTo(
                Request().Url(url=url, defaultTimeout=int(self.get_timeout(timeout)))
            )
            logger.info(response.log)

    def _get_screenshot_path(self, filename: str, fileType: str) -> Path:
        if os.path.isabs(filename):
            d, file = os.path.split(filename)
            directory = Path(d)
            filename = os.path.splitext(file)[0]
        else:
            directory = self.screenshots_output
        # Filename didn't contain {index}
        if "{index}" not in filename:
            return directory / filename
        index = 0
        while True:
            index += 1
            indexed = Path(filename.replace("{index}", str(index)))
            logger.trace(indexed)
            path = directory / indexed
            # Unique path was found
            if not path.with_suffix(f".{fileType}").is_file():
                return path

    @keyword(tags=("PageContent",))
    def take_screenshot(
        self,
        filename: str = "robotframework-browser-screenshot-{index}",
        selector: Optional[str] = None,
        fullPage: bool = False,
        fileType: ScreenshotFileTypes = ScreenshotFileTypes.png,
        quality: Optional[int] = None,
        timeout: Optional[timedelta] = None,
        crop: Optional[BoundingBox] = None,
        disableAnimations: bool = False,
        mask: Union[List[str], str] = "",
        omitBackground: bool = False,
    ) -> str:
        """Takes a screenshot of the current window or element and saves it to disk.

        ``filename`` Filename into which to save. The file will be saved into the robot framework
         ${OUTPUTDIR}/browser/screenshot directory by default, but it can overwritten by providing
         custom path or filename. String ``{index}`` in filename will be replaced with a rolling
         number. Use this to not override filenames. If filename equals to EMBED (case insensitive),
         then screenshot is embedded as Base64 image to the log.html. The image is saved temporally
         to the disk and warning is displayed if removing the temporary file fails.

         The ${OUTPUTDIR}/browser/ is removed at the first suite startup.

        ``selector`` Take a screenshot of the element matched by selector.
        See the `Finding elements` section for details about the selectors.
        If not provided take a screenshot of current viewport.

        ``fullPage`` When True, takes a screenshot of the full scrollable page,
        instead of the currently visible viewport. Defaults to False.

        ``fileType`` <"png"|"jpeg"> Specify screenshot type, defaults to png.

        ``quality`` The quality of the image, between 0-100. Not applicable to png images.

        ``timeout`` Maximum time how long taking screenshot can last, defaults to library timeout.
        Supports Robot Framework time format, like 10s or 1 min, pass 0 to disable timeout.
        The default value can be changed by using the `Set Browser Timeout` keyword.

        ``crop`` Crops the taken screenshot to the given box. It takes same dictionary as returned from `Get BoundingBox`.
        Cropping only works on page screenshot, so if no selector is given.

        ``disableAnimations`` When set to ``True``, stops CSS animations, CSS transitions and Web Animations.
        Animations get different treatment depending on their duration:
         - finite animations are fast-forwarded to completion, so they'll fire transitionend event.
         - infinite animations are canceled to initial state, and then played over after the screenshot.

        ``mask`` Specify selectors that should be masked when the screenshot is taken.
        Masked elements will be overlayed with a pink box ``#FF00FF`` that completely covers its bounding box.
        Arguemnt can take a single selector string or a list of selector strings if multiple different elements should be masked.

        ``omitBackground`` Hides default white background and allows capturing screenshots with transparency.
        Not applicable to jpeg images.

        Keyword uses strict mode if selector is defined. See `Finding elements` for more details
        about strict mode.

        Example
        | `Take Screenshot`                                 # Takes screenshot from page with default filename
        | `Take Screenshot`   selector=id=username_field    # Captures element in image
        | # Takes screenshot with jpeg extension, defines image quality and timeout how long taking screenhost should last
        | `Take Screenshot`   fullPage=True    fileType=jpeg    quality=50    timeout=10s
        """
        if self._is_embed(filename):
            logger.debug("Embedding image to log.html.")
        else:
            logger.debug(f"Using {filename} to take screenshot.")
        string_path_no_extension = str(
            self._get_screenshot_path(filename, fileType.name)
        )
        with self.playwright.grpc_channel() as stub:
            options = {
                "path": f"{string_path_no_extension}.{fileType.name}",
                "fileType": fileType.name,
                "fullPage": fullPage,
                "timeout": int(self.get_timeout(timeout)),
                "omitBackground": omitBackground,
            }
            if mask:
                if isinstance(mask, str):
                    mask_selectors = [mask]
                elif isinstance(mask, Iterable):
                    mask_selectors = [str(s) for s in mask]
                else:
                    raise ValueError(
                        f"'mask' argument is neither string nor list of string. It is {type(mask)}"
                    )
            else:
                mask_selectors = None

            if quality is not None:
                options["quality"] = max(min(100, quality), 0)
            if disableAnimations:
                options["animations"] = "disabled"
            if crop:
                options["clip"] = crop
            response = stub.TakeScreenshot(
                Request().ScreenshotOptions(
                    selector=selector or "",
                    mask=json.dumps(mask_selectors),
                    options=json.dumps(options),
                    strict=self.strict_mode,
                )
            )
            logger.debug(response.log)
            file_path = response.body

            if self._is_embed(filename):
                return self._embed_to_log(file_path)
            return self._log_image_link(file_path)

    def _log_image_link(self, file_path: str) -> str:
        relative_path = get_link_path(file_path, self.outputdir)
        logger.info(
            '</td></tr><tr><td colspan="3">'
            f'<a href="{relative_path}"><img src="{relative_path}" width="800px"></a>',
            html=True,
        )
        return file_path

    def _embed_to_log(self, file_path) -> str:
        png = Path(file_path)
        with png.open("rb") as png_file:
            encoded_string = base64.b64encode(png_file.read())
        # log statement is copied from:
        # https://github.com/robotframework/SeleniumLibrary/blob/master/src/SeleniumLibrary/keywords/screenshot.py
        logger.info(
            '</td></tr><tr><td colspan="3">'
            '<img alt="screenshot" class="robot-seleniumlibrary-screenshot" '
            f'src="data:image/png;base64,{encoded_string.decode()}" width="900px">',
            html=True,
        )
        try:
            png.unlink()
        except Exception:
            logger.warn(f"Could not remove {png}")
        return "EMBED"

    def _is_embed(self, filename: str) -> bool:
        return True if filename.upper() == "EMBED" else False

    @keyword(tags=("Setter", "Config"))
    def set_browser_timeout(self, timeout: timedelta) -> str:
        """Sets the timeout used by most input and getter keywords.

        ``timeout`` Timeout of it is for current playwright context
        and for new contexts. Supports Robot Framework
        [https://robotframework.org/robotframework/latest/RobotFrameworkUserGuide.html#time-format|time format]
        . Returns the previous value of the timeout.

        Example:
        | ${old_timeout} =    `Set Browser Timeout`    1m 30 seconds
        | Click     //button
        | `Set Browser Timeout`    ${old_timeout}
        """
        old_timeout = self.millisecs_to_timestr(self.timeout)
        self.timeout = self.convert_timeout(timeout)
        try:
            with self.playwright.grpc_channel() as stub:
                response = stub.SetTimeout(Request().Timeout(timeout=self.timeout))
                logger.info(response.log)
        except Exception as error:  # Suppress  all errors
            if "Browser has been closed" in str(error):
                logger.debug(f"Suppress error {error} when setting timeout.")
            else:
                raise
        return old_timeout

    @keyword(tags=("Setter", "Config"))
    def set_retry_assertions_for(self, timeout: timedelta) -> str:
        """Sets the timeout used in retrying assertions when they fail.

        `Set Browser timeout` controls how long Playwright will perform
        waiting in the node side, assertion retry will determine how long
        Browser library will retry the playwright operation. Generally
        assertion timeout should be longer than the timeout set by
        `Set Browser timeout`.

        Returns the previous value of the assertion retry.

        Example:
        | `Set Browser Timeout`    10 seconds
        | ${old} =    `Set Retry Assertions For`    30s
        | `Get Title`    ==    Login Page
        | `Set Retry Assertions For`    ${old}

        Example waits 10 seconds on Playwright to get the page title and library
        will retry 30 seconds to make sure that title is correct.
        """
        old_retry_assertions_for = self.millisecs_to_timestr(self.retry_assertions_for)
        self.retry_assertions_for = self.convert_timeout(timeout)
        return old_retry_assertions_for

    @keyword(tags=("Setter", "BrowserControl"))
    def set_viewport_size(self, width: int, height: int):
        """Sets current Pages viewport size to specified dimensions.

        In the case of multiple pages in a single browser,
        each page can have its own viewport size. However,
        `New Context` allows to set viewport size (and more) for all
        later opened pages in the context at once.

        `Set Viewport Size` will resize the page.
        A lot of websites don't expect phones to change size,
        so you should set the viewport size before navigating to
        the page with `New Context` before opening the page itself.

        ``width`` Sets the width size.

        ``height`` Sets the height size.
        """
        with self.playwright.grpc_channel() as stub:
            response = stub.SetViewportSize(
                Request().Viewport(height=height, width=width)
            )
            logger.info(response.log)

    @keyword(tags=("Setter", "BrowserControl"))
    def set_offline(self, offline: bool = True):
        """Toggles current Context's offline emulation.

        ``offline`` Toggles the offline mode. Set to False to switch back
        to online mode. Defaults to True.
        """
        with self.playwright.grpc_channel() as stub:
            response = stub.SetOffline(Request().Bool(value=offline))
            logger.info(response.log)

    @keyword(tags=("Setter", "BrowserControl"))
    def set_geolocation(
        self, latitude: float, longitude: float, accuracy: Optional[float] = None
    ):
        """Updated the correct Context's geolocation.

        Latitude can be between -90 and 90 and longitude can be between -180 and 180.
        Accuracy of the location must be positive number and defaults to 0. When
        creating context, grant ``geolocation`` permission for pages to read its geolocation.

        Example:
        | ${permissions} =    Create List    geolocation
        | `New Context`    permissions=${permissions}
        | `Set Geolocation`    60.173708, 24.982263    3    # Points to Korkeasaari in Helsinki.
        """
        geolocation_dict = {"latitude": latitude, "longitude": longitude}
        if accuracy:
            geolocation_dict["accuracy"] = accuracy
        geolocation = json.dumps(geolocation_dict)
        logger.info(geolocation)
        with self.playwright.grpc_channel() as stub:
            response = stub.SetGeolocation(Request().Json(body=geolocation))
            logger.info(response.log)

    @keyword(tags=("Setter", "BrowserControl"))
    def reload(self):
        """Reloads current active page."""
        with self.playwright.grpc_channel() as stub:
            response = stub.Reload(Request().Empty())
            logger.info(response.log)

    @keyword(tags=("Setter", "BrowserControl"))
    def grant_permissions(self, *permissions: Permission, origin: Optional[str] = None):
        """Grants permissions to the current context.

        ```permissions``` is a list of permissions to grant.
        Permissions can be one of the following:
        geolocation, notifications, camera, microphone,

        Example:
        | `New Context`
        | `Grant Permissions`    geolocation
        """
        with self.playwright.grpc_channel() as stub:
            response = stub.GrantPermissions(
                Request().Permissions(
                    permissions=[p.name for p in permissions], origin=origin or ""
                )
            )
            logger.info(response.log)

    @keyword(tags=("Setter", "BrowserControl"))
    def clear_permissions(self):
        """Clears all permissions from the current context."""
        with self.playwright.grpc_channel() as stub:
            response = stub.ClearPermissions(Request().Empty())
            logger.info(response.log)
